from django.test import TestCase
from django.contrib.auth.models import User, Group

from projector.forms import DashboardAddMemberForm

class DashboardAddMemberFormTest(TestCase):

    def setUp(self):
        self.group = Group.objects.create(name='admins')
        self.user = User.objects.create(username='admin')
        self.user.groups.add(self.group)
        profile = self.user.get_profile()
        profile.group = self.group
        profile.is_team = True
        profile.save()

    def test_wrong_user(self):
        data = {'user': 'not-existing-user-name'}
        form = DashboardAddMemberForm(self.group, data)
        self.assertFalse(form.is_valid())
        self.assertTrue('user' in form._errors)

    def test_wrong_username(self):
        wrong_usernames = (' ', '.', '*', 'joe!', '###', ',.<>')
        for username in wrong_usernames:
            data = {'user': username}
            form = DashboardAddMemberForm(self.group, data)
            self.assertFalse(form.is_valid())
            self.assertTrue('user' in form._errors)

    def test_proper_user(self):
        joe = User.objects.create(username='joe')
        data = {'user': joe.username}
        form = DashboardAddMemberForm(self.group, data)
        self.assertTrue(form.is_valid())

    def test_already_in_group(self):
        data = {'user': self.user.username}
        form = DashboardAddMemberForm(self.group, data)
        self.assertFalse(form.is_valid())
        self.assertTrue('user' in form._errors)




from django.template.loader import render_to_string
from django.utils.translation import ugettext as _
from django.http import Http404

from projector.models import Membership, Watchable
from vcs.utils.annotate import annotate_highlight
from vcs.exceptions import VCSError
from native_tags.decorators import function, filter

def get_project_permissions(project=None, for_bit=None, user=None):
    """
    Usage::

       {% get_project_permissions PROJECT for USER [as "my_var"] %}

    Example::

       {% get_project_permissions project for request.user as "user_permissions" %}
    """
    assert project and for_bit == "for" and user
    try:
        return Membership.objects.get(project=project, member=user).all_perms
    except Membership.DoesNotExist:
        return []
get_project_permissions.function = True

def put_username_into_url(value, user):
    """
    Puts username into url for the given User object.
    """
    if not user.is_authenticated():
        return value
    prefixes = ['http://', 'https://', 'ftp://']
    for prefix in prefixes:
        if value.lower().startswith(prefix):
            value = ''.join((
                value[:len(prefix)], user.username, '@', value[len(prefix):]))
    return value
put_username_into_url = filter(put_username_into_url)

def is_watched(watchable, by_bit=None, user=None):
    """
    Usage::

       {% is_watched WATCHABLE by USER [as "my_var"] %}

    Provided context contains ``task`` and ``user`` objects we can use it as
    follows::

       {% is_watched task by user %}
    """
    assert watchable and isinstance(watchable, Watchable)
    assert by_bit == 'by'
    assert user
    return watchable.is_watched(user)
is_watched.function = True

def annotate_content(context, filenode, **options):
    """
    Usage::

       {% annotate_content filenode cssclass="code-highlight" %}
    """
    def annotate_changeset(changeset):
        template_name = "projector/project/repository/"\
                        "annotate_changeset_cell.html"

        context['line_changeset'] = changeset
        out = render_to_string(template_name, context)
        return out

    order = ['annotate', 'ls', 'code']
    headers = {}
    try:
        return annotate_highlight(filenode, order=order, headers=headers,
            annotate_from_changeset_func=annotate_changeset, **options)
    except VCSError:
        raise Http404
annotate_content = function(annotate_content, is_safe=True, takes_context=True)




import projector
from django.test import TestCase

class ProjectorVersionTests(TestCase):
    """
    Checks if ``__version__`` and ``VERSION`` attributes
    are set correctly in main module.
    """

    def test_version(self):
        self.assertTrue(isinstance(projector.__version__, str))
        self.assertTrue(isinstance(projector.VERSION, tuple))

        VERSION_LENGTH = len(projector.VERSION)
        self.assertTrue(3 <= VERSION_LENGTH and VERSION_LENGTH <= 5)




from django.contrib.auth.models import User, Group
from django.contrib import messages
from django.contrib.sites.models import Site
from django.utils.translation import ugettext as _
from django.shortcuts import get_object_or_404, redirect
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse

from projector.core.controllers import View
from projector.forms import UserProfileForm, UserConvertToTeamForm
from projector.forms import ExternalForkWizard, ExternalForkSourcesForm
from projector.forms import DashboardAddMemberForm
from projector.models import Project
from projector.settings import get_config_value


def can_fork_external():
    """
    External fork is available if ``FORK_EXTERNAL_MAP`` is set to ``True``
    **AND** ``FORK_EXTERNAL_MAP`` is not empty dict.
    """
    if get_config_value('FORK_EXTERNAL_ENABLED') is True and \
        get_config_value('FORK_EXTERNAL_MAP'):
        return True
    return False


class UserListView(View):
    template_name = 'projector/accounts/user_list.html'

    def response(self, request):
        user_list = User.objects.all()
        context = {
            'user_list': user_list,
        }
        return context


class UserHomepageView(View):
    template_name = 'projector/accounts/user_homepage.html'

    def response(self, request):
        if not request.user.is_authenticated():
            msg = _("You are not logged in. You may view only public content "
                "of this website")
            messages.warning(request, msg)
            return {}
        # Authed user homepage
        context = {
            'profile': request.user.get_profile(),
            'owned_task_list': request.user.owned_task\
                .select_related('status', 'project', 'priority',
                    'milestone', 'component')\
                .filter(status__is_resolved=False)\
                .order_by('project')
        }
        return context


class UserProfileDetailView(View):
    """
    Public profile of the given user.
    """
    template_name = 'projector/accounts/profile.html'

    def response(self, request, username):
        user = get_object_or_404(User, username=username)
        context = {
            'profile': user.get_profile(),
            'project_list': Project.objects.for_member(user, request.user),
            'groups': Group.objects.filter(user__userprofile__is_team=True)\
                .filter(user=user)
        }
        return context


class UserDashboardView(View):
    """
    Edit profile view.
    """

    template_name = 'projector/accounts/dashboard.html'

    def response(self, request):
        if request.user.is_anonymous() or not request.user.is_active:
            raise PermissionDenied
        form = UserProfileForm(request.POST or None,
            instance=request.user.get_profile())
        if request.method == 'POST' and form.is_valid():
            form.save()
            message = _("Profile updated successfully")
            messages.success(request, message)
            return redirect(reverse('projector_users_profile_detail',
                kwargs={'username': request.user.username}))
        self.context['form'] = form
        self.context['profile'] = form.instance
        self.context['site'] = Site.objects.get_current()
        self.context['can_fork_external'] = can_fork_external()
        return self.context


class UserDashboardForkView(View):
    """
    Fork external project view.
    """

    template_name = 'projector/accounts/dashboard_fork.html'
    login_required = True

    def response(self, request):
        if get_config_value('FORK_EXTERNAL_ENABLED') is not True:
            messages.warning(request, _("External forks are disabled"))
            return redirect(reverse('projector_dashboard'))
        elif not get_config_value('FORK_EXTERNAL_MAP'):
            messages.warning(request, _("No external fork forms available"))
            return redirect(reverse('projector_dashboard'))

        # Zero at forms list parameter of wizard is only a placeholder - we
        # change form_list dynamically at wizard's ``process_step`` method
        wizard = ExternalForkWizard([ExternalForkSourcesForm, 0])
        wizard.extra_context['profile'] = request.user.get_profile()
        return wizard(request)


class UserDashboardConvert2TeamView(View):
    """
    Convert to team view.
    """

    template_name = 'projector/accounts/dashboard-convert-confirm.html'

    def response(self, request):
        if request.user.is_anonymous() or not request.user.is_active:
            raise PermissionDenied
        if request.user.get_profile().is_team:
            messages.warning(request,
                _("This account have been already converted to team!"))
            return redirect(reverse('projector_users_profile_detail',
                kwargs={'username': request.user.username}))
        form = UserConvertToTeamForm(request.POST or None)
        form.user = request.user
        if request.method == 'POST':
            if form.is_valid():
                msg = _("You have successfully converted account into Team")
                messages.success(request, msg)
                return redirect(reverse('projector_users_profile_detail',
                    kwargs={'username': request.user.username}))
            else:
                msg = _("Errors occured during account to team conversion")
                messages.error(request, msg)
        self.context['form'] = form
        self.context['profile'] = request.user.get_profile()
        return self.context

class UserDashboardAddMember(View):
    """
    Add new member view.
    """

    template_name = 'projector/accounts/dashboard-add-new-member.html'

    def response(self, request):
        profile = request.user.get_profile()
        if request.user.is_anonymous() or not request.user.is_anonymous or \
            not profile or not profile.is_team:
            messages.warning(request, _('Only teams are allowed to add member'))
            return redirect(reverse('projector_users_profile_detail',
                kwargs={'username': request.user.username}))
        form = DashboardAddMemberForm(profile.group, request.POST or None)
        if request.method == 'POST' and form.is_valid():
            form.save()
            user = form.cleaned_data['user']
            msg = _("User %s is now member of this team!" % user)
            messages.success(request, msg)
            return redirect(reverse('projector_users_profile_detail',
                kwargs={'username': request.user.username}))
        self.context['form'] = form
        self.context['profile'] = profile
        return self.context




import logging
import datetime

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import Count
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render_to_response
from django.template import RequestContext
from django.utils.simplejson import dumps
from django.utils.translation import ugettext as _
from django.utils.decorators import method_decorator

from projector.core.controllers import View
from projector.core.exceptions import ProjectorError
from projector.models import Project, State
from projector.forms import ProjectCreateForm, ProjectEditForm, ConfigForm,\
    ProjectForkForm
from projector.settings import get_config_value
from projector.signals import setup_project

from vcs.web.simplevcs import settings as simplevcs_settings
from vcs.web.simplevcs.utils import get_mercurial_response, is_mercurial
from vcs.web.simplevcs.utils import log_error, basic_auth, ask_basic_auth
from vcs.web.simplevcs.exceptions import NotMercurialRequest

login_required_m = method_decorator(login_required)

class ProjectView(View):
    """
    Base class for all projector views.

    Logic should be implemented at ``response`` method.

    Would check necessary permissions defined by class attributes: ``perms``,
    ``perms_POST`` and ``perms_POST``. ``perms`` are always checked,
    ``perms_POST`` are additional checks which would be made for ``GET`` method
    requests only and ``perms_POST`` would be made for ``POST`` method requests.
    ``perms_private`` would be checked for private projects only.

    """

    perms = []
    perms_private = ['view_project']
    perms_GET = []
    perms_POST = []

    template_error_name = 'projector/project/error.html'
    template_pending_name = 'projector/project/pending.html'

    def __init__(self, request, username=None, project_slug=None, *args,
            **kwargs):
        self.request = request
        self.project = get_object_or_404(Project, slug=project_slug,
            author__username=username)
        self.author = self.project.author
        self.check_permissions()
        super(ProjectView, self).__init__(request=request, username=username,
            project_slug=project_slug, *args, **kwargs)
        self.context['project'] = self.project
        self.context['STATES'] = State

        # Set forks without additional queries if is a root (not forked project)
        if self.project.is_root():
            self.context['forks'] = [self.project]
            self.context['project_root'] = self.project
            if request.user.is_anonymous():
                self.context['user_fork'] = None
            elif request.user == self.project.author:
                self.context['user_fork'] = self.project
            else:
                self.context['user_fork'] = self.project.get_fork_for_user(
                    request.user)
        else:
            forks = self.project.get_all_forks()
            self.context['forks'] = forks
            self.context['project_root'] = self.project.get_root()
            user_fork = None
            if request.user.is_anonymous():
                self.context['user_fork'] = None
            else:
                for fork in forks:
                    if fork.author_id == request.user.id:
                        user_fork = fork
                        break
            self.context['user_fork'] = user_fork


    def __after__(self):
        if self.project.state == State.ERROR:
            return render_to_response(self.template_error_name, self.context,
                RequestContext(self.request))
        if self.project.is_pending():
            return render_to_response(self.template_pending_name, self.context,
                RequestContext(self.request))

    def get_required_perms(self):
        """
        Returns list of required perms based on instance's attributes.
        If modified, make sure it's thread-safe, i.e. don't change self.perms
        directly but create new temporary list of permissions.
        """
        perms = [p for p in self.perms]

        if self.request.method == 'GET':
            perms += [p for p in self.perms_GET]
        elif self.request.method == 'POST':
            perms += [p for p in self.perms_POST]

        if self.project.is_private():
            perms += [p for p in self.perms_private]
        perms = set(perms)
        return perms

    def check_permissions(self):
        # Owner's are always allowed to do anything with their projects
        # - this would make less database hits
        if self.project.author == self.request.user:
            return
        perms = self.get_required_perms()
        for perm in perms:
            if not self.request.user.has_perm(perm, self.project):
                if settings.DEBUG:
                    logging.debug("User %s has no permission %s for project %s"
                        % (self.request.user, perm, self.project))
                raise PermissionDenied()


class ProjectState(ProjectView):

    def __after__(self):
        pass

    def response(self, request, username, project_slug):
        data = {'state': self.project.state}
        if self.project.state == State.ERROR:
            data['error_text'] = self.project.error_text or ''
        json_data = dumps(data)
        return HttpResponse(json_data, content_type='application/json')


class ProjectDetailView(ProjectView):
    """
    Returns selected project's detail for user given in ``request``.
    We make necessary permission checks *after* dispatching between
    normal and mercurial request, as mercurial requests has it's own
    permission requirements.
    """

    template_name = 'projector/project/detail.html'
    csrf_exempt = True

    def get_required_perms(self):
        if is_mercurial(self.request):
            # For mercurial requests lets undelying view to
            # manage permissions checking
            return []
        return super(ProjectDetailView, self).get_required_perms()

    def response(self, request, username, project_slug):
        try:

            if is_mercurial(request):
                return self.response_hg(request, self.project)
            last_part = request.path.split('/')[-1]
            if last_part and last_part != project_slug:
                raise Http404("Not a mercurial request and path longer than "
                    " should be: %s" % request.path)

            # project is injected into the context at ProjectView constructor
            # so we do not need to add it here
            return self.context

        except Exception, err:
            dont_log_exceptions = (PermissionDenied,)
            if not isinstance(err, dont_log_exceptions):
                log_error(err)
            raise err

    def response_hg(self, request, project):
        return _project_detail_hg(request, project)


def _project_detail_hg(request, project):
    """
    Wrapper for vcs.web.simplevcs.views.hgserve view as before we go any further
    we need to check permissions.
    TODO: Should use higher level simplevcs method
    """
    if not is_mercurial(request):
        msg = "_project_detail_hg called for non mercurial request"
        logging.error(msg)
        raise NotMercurialRequest(msg)

    if request.method not in ('GET', 'POST'):
        raise NotMercurialRequest("Only GET/POST methods are allowed, got %s"
            % request.method)
    # Allow to read from public projects
    if project.is_public() and request.method == 'GET':
        mercurial_info = {
            'repo_path': project._get_repo_path(),
            'push_ssl': simplevcs_settings.PUSH_SSL,
        }
        return get_mercurial_response(request, **mercurial_info)

    # Check if user have been already authorized or ask to
    request.user = basic_auth(request)
    if request.user is None:
        return ask_basic_auth(request,
            realm=project.config.basic_realm)

    if project.is_private() and request.method == 'GET' and\
        not request.user.has_perm('can_read_repository', project):
        raise PermissionDenied("User %s cannot read repository for "
            "project %s" % (request.user, project))
    elif request.method == 'POST' and\
        not request.user.has_perm('can_write_to_repository',project):
        raise PermissionDenied("User %s cannot write to repository "
            "for project %s" % (request.user, project))

    mercurial_info = {
        'repo_path': project._get_repo_path(),
        'push_ssl': simplevcs_settings.PUSH_SSL,
    }

    if request.user and request.user.is_active:
        mercurial_info['allow_push'] = request.user.username

    response = get_mercurial_response(request, **mercurial_info)
    return response


class ProjectListView(View):
    """
    Project listing view.
    """

    template_name = 'projector/project/list.html'

    def response(self, request):
        project_list = Project.objects.for_user(user=request.user)\
            .annotate(Count('task', distinct=True))
        context = {
            'project_list' : project_list,
        }
        return context


class ProjectCreateView(View):
    """
    New project creation view.
    """

    template_name = 'projector/project/create.html'

    @login_required_m
    def response(self, request, username=None):
        # TODO: what with username param? should it be required?
        # it's not used for now...
        project = Project(
            author=request.user,
        )
        form = ProjectCreateForm(request.POST or None, instance=project,
            initial={'public': u'private'})
        if request.method == 'POST' and form.is_valid() and \
                self.can_create(request.user, request):
            project = form.save()
            return redirect(project.get_absolute_url())
        context = {
            'form' : form,
        }
        return context

    @staticmethod
    def can_create(user, request=None):
        """
        Checks if given user can create project. Note that this function will
        not validate project itself. If
        :setting:`PROJECTOR_MILIS_BETWEEN_PROJECT_CREATION` is greater than
        miliseconds since last time this user has created a project then he or
        she is allowed to create new one.

        If user is trying to create more project than specified by
        MAX_PROJECTS_PER_USER configuration value then we disallow.

        If request is given, send messages.
        """

        def send_error(request, message):
            if request:
                messages.error(request, message)

        try:
            date = Project.objects.filter(author=user)\
                .only('name', 'created_at')\
                .order_by('-created_at')[0].created_at
            delta = datetime.datetime.now() - date
            milis = delta.seconds * 1000
            need_to_wait = get_config_value('MILIS_BETWEEN_PROJECT_CREATION')\
                - milis
            need_to_wait /= 1000
            if need_to_wait > 0:
                send_error(request, _("You would be allowed to create a new "
                    "project in %s seconds" % need_to_wait))
                return False
        except IndexError:
            pass
        count = user.project_set.count()
        too_many = count - get_config_value('MAX_PROJECTS_PER_USER')
        if too_many > 0:
            send_error(request, _("You cannot create more projects"))
            return False
        return True


class ProjectForkView(ProjectView):
    """
    Project fork (internal fork) view.
    """

    template_name = 'projector/project/fork.html'

    @login_required_m
    def response(self, request, username, project_slug):
        user_fork = self.project.get_fork_for_user(request.user)
        if user_fork:
            messages.warning(request, _("User has already forked this project"))
            return redirect(user_fork.get_absolute_url())

        form = ProjectForkForm(request.POST or None)
        if request.method == 'POST':
            if form.is_valid():
                # TODO: Move can_create from ProjectCreateView class
                ProjectCreateView.can_create(request.user, request)
                try:
                    fork = self.project.fork(user=request.user)
                    setup_project.send(sender=Project, instance=fork)
                    return redirect(fork.get_absolute_url())
                except ProjectorError, err:
                    messages.error(request, str(err))
                return redirect(self.project.get_absolute_url())
        self.context['form'] = form
        return self.context


class ProjectEditView(ProjectView):
    """
    Update project view.
    """

    template_name = 'projector/project/edit.html'
    perms = ['view_project', 'change_project', 'admin_project']

    def validate_project_form(self, request):
        public_val = self.project.is_public() and u'public' or u'private'
        if request.method == 'POST' and \
                request.POST.get('submit_project', False):
            form = ProjectEditForm(request.POST, instance=self.project,
                initial={'public': public_val})
            if form.is_valid():
                msg = _("Project edited successfully")
                self.project = form.save()
                messages.success(request, msg)
                return redirect(self.project.get_edit_url())
            else:
                msg = _("Form has not validated")
                messages.error(request, msg)
        else:
            form = ProjectEditForm(instance=self.project,
                initial={'public': public_val})
        self.context['form'] = form

    def validate_config_form(self, request):
        if request.method == 'POST' and \
                request.POST.get('submit_config', False):
            form = ConfigForm(request.POST, instance=self.project.config)
            if form.is_valid():
                form.instance.editor = request.user
                form.save()
                msg = _("Project's configuration updated successfully")
                messages.success(request, msg)
            else:
                msg = _("Errors occured while trying to update "
                            "project's configuration")
                messages.error(request, msg)
        else:
            form = ConfigForm(instance=self.project.config)
        self.context['form_config'] = form

    @login_required_m
    def response(self, request, username, project_slug):
        self.context['project'] = self.project
        result = self.validate_project_form(request)
        if result:
            return result
        result = self.validate_config_form(request)
        if result:
            return result
        return self.context





import urlparse

from django.contrib.auth.models import User, Group
from django.core.urlresolvers import reverse
from django.test.client import Client

from projector.tests.base import ProjectorTestCase
from projector.models import Project, Membership, Team

from guardian.shortcuts import assign

class ProjectorPermissionTests(ProjectorTestCase):

    fixtures = ['test_data.json']

    def setUp(self):
        self.client = Client()

        self.admin = User.objects.get(username='admin')
        self.admin._plain_password = 'admin'

        self.jack = User.objects.get(username='jack')
        self.jack._plain_password = 'jack'

        self.joe = User.objects.get(username='joe')
        self.joe._plain_password = 'joe'

        self.noperms = User.objects.get(username='noperms')
        self.noperms._plain_password = 'noperms'

        # Create projects
        self.public_project = Project.objects.get(name='public project')
        self.private_project = Project.objects.get(name='private project')

    def test_anonymous(self):
        # Test public project for anonymous user
        self.client.logout()

        url = self.public_project.get_absolute_url()
        self._get_response(url)

        url = reverse('projector_project_create')
        response = self._get_response(url, follow=True)
        # Redirects directly to login page with next parameter
        self.assertTrue(len(response.redirect_chain) == 1)
        passed_url, passed_code = response.redirect_chain[0]
        parsed = urlparse.urlparse(passed_url)
        self.assertEqual(parsed.path, reverse('auth_login'))
        self.assertEqual(parsed.query, 'next=%s' % url)

    def test(self):

        # =================== #
        # Test public project #
        # =================== #

        for user in (self.admin, self.noperms, self.joe, self.jack):
            client = self.client
            client.login(username=user.username, password=user._plain_password)
            url = self.public_project.get_absolute_url()
            response = client.get(url)
            self.assertTrue(response.status_code == 200,
                "User %s doesn't have permission to view public project!"
                % user)
            client.logout()

        # ==================== #
        # Test private project #
        # ==================== #

        user = self.jack
        client = self.client
        client.login(username=user.username, password=user._plain_password)

        urls = (
            self.private_project.get_absolute_url(),
            self.private_project.get_edit_url(),
            self.private_project.get_members_url(),
            self.private_project.get_members_add_url(),
            #self.private_project.get_members_edit_url(),
            self.private_project.get_teams_url(),
            self.private_project.get_teams_add_url(),
            #self.private_project.get_teams_edit_url(),
            self.private_project.get_fork_url(),
            self.private_project.get_state_url(),

            self.private_project.get_task_list_url(),
            self.private_project.get_create_task_url(),
            self.private_project.get_milestones_url(),
            self.private_project.get_milestone_add_url(),
            self.private_project.get_components_url(),
            self.private_project.get_component_add_url(),
            self.private_project.get_workflow_url(),
            self.private_project.get_workflow_edit_url(),
            self.private_project.get_workflow_add_status_url(),
            self.private_project.get_browse_repo_url(),
            self.private_project.get_changesets_url(),
        )
        for url in urls:
            response = self.client.get(url)
            self.assertTrue(response.status_code == 403,
                "User %s tried to access %s but returned code was %s"
                % (user, url, response.status_code))

        # Now we add member and a team to the project and check if
        # they cannot be edited
        new_member = User.objects.create_user(
            username='new_member',
            email='new_member@non-existence-1231231.com',
            password='new_member')
        new_member._plain_password = 'new_member'

        Membership.objects.create(member=new_member,
            project=self.private_project)

        new_group = Group.objects.create(name='new_team')
        team = Team.objects.create(group=new_group,
            project=self.private_project)

        responses = (
            self.client.get(
                self.private_project.get_members_edit_url(new_member.username)),
            self.client.post(
                self.private_project.get_members_edit_url(new_member.username),
                data = {}),
            self.client.get(
                self.private_project.get_teams_edit_url(team.group.name)),
            self.client.post(
                self.private_project.get_teams_edit_url(team.group.name),
                data = {}),
        )
        for response in responses:
            self.assertTrue(response.status_code == 403,
                "User %s tried to access %s %s but returned code was %s"
                % (user, response.request['REQUEST_METHOD'],
                    response.request['PATH_INFO'], response.status_code))

        # Test project listing
        url = reverse('projector_project_list')
        resp = self._get_response(url)

        project_list = resp.context['project_list']
        for_user = Project.objects.for_user(user)
        self.assertEqual(set(project_list), set(for_user),
            "Project list retrieved from page (%s) is not equal with list for "
            "user (%s)" % (project_list, for_user))

        # Now we make jack a member but without permissions he shouldn't be
        # able to see anything new
        Membership.objects.create(member=user, project=self.private_project)
        for url in urls:
            response = self.client.get(url)
            self.assertTrue(response.status_code == 403,
                "User %s tried to access %s but returned code was %s"
                % (user, url, response.status_code))

        responses = (
            self.client.get(
                self.private_project.get_members_edit_url(new_member.username)),
            self.client.post(
                self.private_project.get_members_edit_url(new_member.username),
                data = {}),
            self.client.get(
                self.private_project.get_teams_edit_url(team.group.name)),
            self.client.post(
                self.private_project.get_teams_edit_url(team.group.name),
                data = {}),
        )
        for response in responses:
            self.assertTrue(response.status_code == 403,
                "User %s tried to access %s %s but returned code was %s"
                % (user, response.request['REQUEST_METHOD'],
                    response.request['PATH_INFO'], response.status_code))

        # We add "view_project" permission for jack
        assign('view_project', user, self.private_project)
        url = self.private_project.get_absolute_url()
        self._get_response(url)

        urls = (
            self.private_project.get_components_url(),
            self.private_project.get_milestones_url(),
            self.private_project.get_workflow_url(),
            self.private_project.get_state_url(),
        )
        for ulr in urls:
            self._get_response(url, code=200)

        self.client.logout()




from django.test.client import Client
from django.contrib.auth.models import User

from projector.tests.base import ProjectorTestCase
from projector.models import Project, Milestone

class MilestoneTest(ProjectorTestCase):

    def setUp(self):
        self.client = Client()
        cred = 'milestoner'
        self.user, created = User.objects.get_or_create(
            username = cred,
            email = 'milestoner@example.com',
            is_superuser = True,
            is_active = True)
        self.user.set_password(cred)
        self.user.save()
        self.user._plain_password = cred
        self.project = Project.objects.create_project(
            name = 'milestone-test-project',
            slug = 'milestone-test-project',
            author = self.user,
        )

    def test_name_uniqueness(self):
        self.client.login(username = self.user.username,
            password = self.user._plain_password)

        deadline = u'2012-01-01'

        new_milestone_name = u'milestone-1'
        response = self._get_response(
            url = self.project.get_milestone_add_url())
        self._get_response(
            url = self.project.get_milestone_add_url(),
            data = {
                'name': new_milestone_name,
                'description': u'anything',
                'deadline': deadline,
            },
            method = 'POST', follow = True)

        self.assertEqual(self.project.milestone_set.count(), 1)

        response = self._get_response(
            url = self.project.get_milestone_add_url(),
            data = {
                'name': new_milestone_name,
                'description': u'anything',
                'deadline': deadline,
            },
            method = 'POST', follow = True)

        # Should not be added as it has same name
        self.assertEqual(self.project.milestone_set.count(), 1)
        name_field_errors = response.context['form']._errors['name']
        self.assertTrue(len(name_field_errors) > 0,
            "Milestone with name '%s' for this project already exist "
            "so we shouldn't be able to add new one" % new_milestone_name)

        another_name = u'milestone-2'
        response = self._get_response(
            url = self.project.get_milestone_add_url(),
            data = {
                'name': another_name,
                'description': u'anything',
                'deadline': deadline,
            },
            method = 'POST', follow = True)
        # Should not allow to update name of the milestone to one already
        # existing
        milestone = Milestone.objects.get(project = self.project,
            name = another_name)
        response = self._get_response(
            url = milestone.get_edit_url(),
            data = {
                'name': new_milestone_name,
                'description': u'anything',
                'deadline': deadline,
            },
            method = 'POST', follow = True)
        name_field_errors = response.context['form']._errors['name']
        self.assertTrue(len(name_field_errors) > 0,
            "Milestone with name '%s' for this project already exist "
            "so we shouldn't be able to change old name '%s' to this one"
            % (new_milestone_name, milestone.name))







import logging

from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import ugettext as _
from django.contrib.admin.util import NestedObjects
from django.contrib import messages
from django.contrib.contenttypes.models import ContentType

from projector.models import Team
from projector.views.project import ProjectView
from projector.forms import TeamForm, ProjectTeamPermissionsForm, TeamDeleteForm

from guardian.shortcuts import get_perms
from guardian.models import GroupObjectPermission

class TeamListView(ProjectView):
    """
    Returns teams view.
    """

    template_name = 'projector/project/teams/home.html'
    perms_private = ['view_project', 'view_teams_project']

    def response(self, request, username, project_slug):
        teams = Team.objects.filter(project=self.project)
        context = {
            'project': self.project,
            'team_list': teams,
        }
        return context

class TeamAddView(ProjectView):
    """
    Adds team for a project.
    """

    template_name = 'projector/project/teams/add.html'
    perms = ['view_project', 'add_team_project']

    def response(self, request, username, project_slug):
        team = Team(
            project = self.project,
        )
        form = TeamForm(request.POST or None, instance=team)

        if request.method == 'POST' and form.is_valid():
            logging.info("Saving team %s for project '%s'"
                % (form.instance.group, form.instance.project))
            form.save()
            return redirect(self.project.get_teams_url())
        elif form.errors:
            logging.error("Form contains errors:\n%s" % form.errors)

        context = {
            'project': form.instance.project,
            'form': form,
        }
        return context

class TeamEditView(ProjectView):
    """
    Manages settings and permissions of project's team.
    """

    template_name = 'projector/project/teams/edit.html'
    perms = ['view_project', 'change_team_project']

    def response(self, request, username, project_slug, name):
        team = get_object_or_404(
            Team.objects.select_related('group', 'project'),
            project__author__username=username,
            project__slug=project_slug, group__name=name)

        group, project = team.group, team.project
        team_perms = get_perms(group, project)

        form = ProjectTeamPermissionsForm(request.POST or None,
            team = team,
            initial_permissions = team_perms,
            request = request)
        if request.method == 'POST':
            if form.is_valid():
                form.save()
                messages.success(request, _("Permissions updated"))
            else:
                messages.error(request,
                    _("Errors occured while processing the form"))
            return redirect(team.get_absolute_url())
        context = {
            'project': self.project,
            'form': form,
            'team': team,
            'team_perms': team_perms,
        }
        return context

class TeamDeleteView(ProjectView):
    """
    Removes Group membership from the project.
    """

    template_name = 'projector/project/teams/delete.html'
    perms = ['view_project', 'can_delete_team']

    def response(self, request, username, project_slug, name):
        team = get_object_or_404(
            Team.objects.select_related('group', 'project'),
            project__author__username=username,
            project__slug=project_slug, group__name=name)

        group, project = team.group, team.project

        collector = NestedObjects()
        team._collect_sub_objects(collector)
        form = TeamDeleteForm(request.POST or None)
        perms_to_delete = GroupObjectPermission.objects.filter(
            group = group,
            content_type = ContentType.objects.get_for_model(project),
            object_id = project.id)

        if request.method == 'POST':
            # Confirm removal
            if form.is_valid():
                msg = _("Team removed")
                messages.success(request, msg)
                team.delete()
                perms_to_delete.delete()
                return redirect(project.get_teams_url())
            else:
                msg = _("Couldn't remove team")
                messages.error(request, msg)

        context = {
            'project': project,
            'team': team,
            'form': form,
            'to_delete': collector.nested(),
            'team_perms': perms_to_delete,
        }
        return context




"""
Utilities and helpers for ``django-projector``.
"""
import os

from django.conf import settings
from projector.utils.basic import str2obj, codename_to_label

__all__ = [
    'abspath',
    'codename_to_label',
    'str2obj',
    'using_projector_profile',
]

abspath = lambda *paths: os.path.abspath(os.path.join(*paths))

def using_projector_profile():
    """
    Returns True if ``AUTH_PROFILE_MODULE`` is set to ``projector.UserProfile``
    and False otherwise.
    """
    auth_profile_module = getattr(settings, 'AUTH_PROFILE_MODULE', False)
    return auth_profile_module == 'projector.UserProfile'




from django.contrib.auth.models import User
from django.test import TestCase
from django.db.models import Q
from django.template.defaultfilters import slugify

from projector.models import Project, Transition, Status

class ProjectorStatusTransition(TestCase):

    def setUp(self):
        name = u'ProjectorStatusTransition'
        self.admin = User.objects.create_superuser(
            username = name,
            email = 'admin@example.com',
            password = 'admin',
        )
        self.project = Project.objects.create_project(
            name=name,
            slug=slugify(name),
            author=self.admin,
        )
        Status.objects.create(name='s1', project=self.project, order=2)
        Status.objects.create(name='s2', project=self.project, order=3)

        Transition.objects.filter(
            Q(source__project=self.project) |
            Q(destination__project=self.project))\
            .delete()

    def test_notransitions(self):
        statuses = self.project.status_set.all()
        for src in statuses:
            for dst in statuses:
                self.assertFalse(src.can_change_to(dst))

    def test_transition(self):
        s1, s2 = self.project.status_set.all()[:2]
        Transition.objects.create(source=s1, destination=s2)
        self.assertTrue(s1.can_change_to(s2))
        self.assertFalse(s2.can_change_to(s1))
        Transition.objects.create(source=s2, destination=s1)
        self.assertTrue(s1.can_change_to(s2))
        self.assertTrue(s2.can_change_to(s1))








"""
This module extends standard Django function views and allows us to use *so
called* class-based-views. Django itself contains many features to power up
class based approach to the topic (for instance
``django.utils.decorators.method_decorator``).

During development of ``django-projector`` we just copied codes much too often
and class-based views allow us to complete many tasks in a much simpler way.
"""
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.http import HttpRequest, HttpResponse
from django.utils.decorators import method_decorator
from django.contrib.auth.decorators import login_required

login_required_m = method_decorator(login_required)

class View(object):
    """
    Main view class. Implemntation is focused around ``__call__`` method and
    *classmethod* ``new``.

    Subclasses should implement ``response`` method but it is not required.

    **Class-level attributes**:

    :attribute login_required:
        Default: ``False``.
        If set to ``True``, ``response`` method would be wrapped with standard
        ``django.contrib.auth.decorators.login_required``.

    :attribute template_name:
        Default: ``'base.html'``
        If ``response`` method returns a ``dict`` then this value is used to
        render context.

    **Instance-available attributes**:

    :attribute self.context:
        Context dictionary. During initialization it is set to empty dict
        (``{}``).

    :attribute self.request:
        ``django.http.HttpRequest`` instance passed in by the url resolver.

    :attribute self.args:
        Positional parameters passed in by url resolver.

    :attribute self.kwargs:
        Named parameters passed in by url resolver.

    **Implementation example**::

        from projector.core.controllers import View
        from projector.models import Project

        class ProjectList(View):

            login_required = True
            template_name = 'project_list.html'

            def response(self, request):
                self.context = Project.objects.all()
                return self.context

        class ProjectDetail(View):

            login_required = True
            template_name = 'project_detail.html'

            def response(self, request, project_id):
                self.context['project'] = get_object_or_404(Project, id=project_id)
                return self.context

    Typically, we would hook such defined views at ``urls.py`` module::

        from django.conf.urls.defaults import *

        urlpatterns = patterns('projector.views',
            (r'^projects/$', 'ProjectList'),
            (r'^projects/(?P<project_id>\d+)/$', 'ProjectDetail'),
        )

    Note that url resolvers pass all parameters to the given functions. Our
    subclasses of :view:`View` class are treated very similar to functions
    as they should return ``HttpResponse`` object after being called.
    """
    login_required = False
    template_name = 'base.html'

    def __new__(cls, request, *args, **kwargs):
        """
        Customized new instance static method.
        """
        # Wraps response method with ``login_required`` decorator
        # if ``login_required`` attribute is set to ``True`` at the class
        if cls.login_required:
            cls.response = login_required_m(cls.response)
        view = cls.new(request, *args, **kwargs)

        # If already a HttpResponse we should propagate it
        if isinstance(view, HttpResponse):
            return view
        return view.__call__()

    @classmethod
    def new(cls, request, *args, **kwargs):
        obj = object.__new__(cls)
        if not isinstance(request, HttpRequest):
            raise TypeError(
            "Class based views requires HttpRequest to be passed as first "
            "argument (got %s)" % request)
        obj.request = request
        obj.args = args
        obj.kwargs = kwargs
        obj.context = {}
        # Call __before__ method
        before = obj.__before__()
        if isinstance(before, HttpResponse):
            return before
        obj.__init__(request, *args, **kwargs)
        return obj

    def __before__(self):
        """
        Method called just after class is created and all passed parameters are
        set on the class, so it is possible to access those attributes.

        If this method would return :py:class:`django.http.HttpResponse`
        instance, it would be propagated.
        """
        pass

    def __after__(self):
        """
        Method called at the final step of the view-class processing. If
        :py:class:`django.http.HttpResponse` instance is returned it is
        propagated to the view handler.

        .. note::
           Adding i.e. ``context`` at this stage is pointless - response
           should have been already prepared and changing ``context`` wouldn't
           change the response. ``__after__`` method should be implemented
           for other purposes, like checking status code of prepared response.
           Response object is available by accessing ``self._response``. This
           attribute is available **only** at this stage.
        """
        pass

    def __init__(self, request, *args, **kwargs):
        """
        May be overridden, for example::

            def __init__(self, request, page_slug):
                self.page_slug = page_slug

        """
        pass

    def __call__(self):
        """
        Should be overridden only for special cases as it runs ``__after__``
        method after ``response`` method is executed. By subclassing it is
        possible to change the order of method calls.
        """
        if self.args and self.kwargs:
            response = self.response(self.request, *self.args, **self.kwargs)
        elif self.args:
            response = self.response(self.request, *self.args)
        elif self.kwargs:
            response = self.response(self.request, **self.kwargs)
        else:
            response = self.response(self.request)

        # If response is a dict, then we have to trasform it into HttpResponse
        if isinstance(response, dict):
            response = render_to_response(self.template_name, response,
                RequestContext(self.request))
        self._response = response

        # Call __after__ method just before we return response
        after = self.__after__()
        if isinstance(after, HttpResponse):
            return after
        del self._response
        return response

    def response(self, request, *args, **kwargs):
        """
        Should be overridden at subclass. It always requires
        ``django.http.HttpRequest`` instance to be passed as first positional
        argument.

        May return a ``dict`` or ``django.http.HttpResponse`` instance. If
        ``dict`` is returned it is treated as *context* and view would try
        to render it using ``self.template_name``.

        :returns: ``dict`` or ``django.http.HttpResponse`` instance
        """
        return self.context




from django.utils.importlib import import_module

def codename_to_label(codename):
    rremove = '_project'
    if codename.endswith(rremove):
        codename = codename[:len(codename)-len(rremove)]
    codename = codename\
        .split('.')[1]\
        .replace('_', ' ')\
        .capitalize()
    return codename

def str2obj(text):
    """
    Returns object pointed by the string. In example::

        >>> from projector.models import Project
        >>> point = 'projector.models.Project'
        >>> obj = str2obj(point)
        >>> obj is Project
        True

    """
    modpath, objname = text.rsplit('.', 1)
    mod = import_module(modpath)
    try:
        obj = getattr(mod, objname)
    except AttributeError:
        raise ImportError("Cannot retrieve object from location %s" % text)
    return obj




import sys
import logging

from celery.decorators import task

from django.utils.translation import ugettext as _

from projector.models import Project, State
from projector.utils import str2obj
from projector.settings import get_config_value

@task
def project_create_repository(instance, vcs_alias=None):
    if get_config_value('CREATE_REPOSITORIES'):
        instance.create_repository(vcs_alias)

@task
def setup_project(instance, vcs_alias=None, workflow=None):
    """
    Creates all necessary related objects like statuses with transitions etc.
    It simply calls setup and we do this here as in a production it would most
    probably be called asynchronously (if
    :setting:`PROJECTOR_CREATE_PROJECT_ASYNCHRONOUSLY` is set to ``True``)

    :param instance: instance of :model:`Project`
    :param vcs_alias: alias of vcs backend
    :param workflow: object or string representing project workflow
    """
    logging.debug("Task setup_project called for instance %s" % instance)
    if isinstance(workflow, str):
        workflow = str2obj(workflow)
    #instance.setup(vcs_alias=vcs_alias, workflow=workflow)

    # Prepare if parametrs are given. Otherwise assume that preparation
    # methods have been called already
    if vcs_alias:
        instance.set_vcs_alias(vcs_alias)
    if workflow:
        instance.set_workflow(workflow)

    current_state = instance.state
    # Fire up sub-actions
    try:
        # Membership & Team
        instance.set_memberships()
        current_state = State.MEMBERSHIPS_CREATED
        Project.objects.filter(pk=instance.pk).update(state=current_state)

        # Author permissions
        instance.set_author_permissions()
        current_state = State.AUTHOR_PERMISSIONS_CREATED
        Project.objects.filter(pk=instance.pk).update(state=current_state)

        instance.create_workflow()
        current_state = State.WORKFLOW_CREATED
        Project.objects.filter(pk=instance.pk).update(state=current_state)

        instance.create_config()
        current_state = State.CONFIG_CREATED
        Project.objects.filter(pk=instance.pk).update(state=current_state)

        if get_config_value('CREATE_REPOSITORIES'):
            # We spawn sub task here as creating repositories is most crucial
            # task durgin project setup
            result = project_create_repository.delay(instance)
            result.wait()
            current_state = State.REPOSITORY_CREATED
            Project.objects.filter(pk=instance.pk).update(state=current_state)

        current_state = State.READY
        Project.objects.filter(pk=instance.pk).update(state=current_state)

    except (MemoryError, KeyboardInterrupt):
        raise
    except Exception:
        try:
            from djangodblog.models import Error
            Error.objects.create_from_exception()
        except ImportError:
            pass
        Project.objects.filter(pk=instance.pk).update(state=State.ERROR)
        user_error_text = _("There were some crazy error during project setup "
                            "process")
        logging.error("Error during project setup. Last state was: %s\n"
                      "Info:\n%s\n" % (current_state, sys.exc_info()))
        Project.objects.filter(pk=instance.pk).update(
            error_text=user_error_text)




from django.conf.urls.defaults import *

from projector.feeds import LatestProjectsFeed

urlpatterns = patterns('projector.views',
    url(r'^$', 'project.ProjectListView', name='projector_home'),
)

urlpatterns += patterns('projector.views.project_category',
    url(r'^categories/$',
        view='project_category_list',
        name='projector_project_category_list'),
    url(r'^categories/create/$',
        view='project_category_create',
        name='projector_project_category_create'),
    url(r'^categories/(?P<project_category_slug>[-\w]+)/$',
        'project_category_detail', name='projector_project_category_detail'),
)

# User dashboard
urlpatterns += patterns('projector.views.users',
    url(r'^dashboard/convert-to-team/$',
        view = 'UserDashboardConvert2TeamView',
        name = 'projector_dashboard_convert_to_team'),

    url(r'^dashboard/external-fork/$',
        view = 'UserDashboardForkView',
        name = 'projector_dashboard_fork'),

    url(r'^dashboard/add-member/$',
        view = 'UserDashboardAddMember',
        name = 'projector_dashboard_add_member'),

    url(r'^dashboard/$',
        view = 'UserDashboardView',
        name = 'projector_dashboard'),
)

urlpatterns += patterns('projector.views.project',
    # Basic
    url(r'^projects/$',
        view='ProjectListView',
        name='projector_project_list'),
    url(r'^projects/create-project/$',
        view='ProjectCreateView',
        name='projector_project_create'),
    url(r'^projects/fork-external-project/$',
        view='ProjectCreateView',
        name='projector_project_fork_external'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/admin/$',
        view='ProjectEditView',
        name='projector_project_edit'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/fork/$',
        view='ProjectForkView',
        name='projector_project_fork'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/state/$',
        view='ProjectState',
        name='projector_project_state'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/$',
        view='ProjectDetailView', name='projector_project_detail'),

)

# Milestones
urlpatterns += patterns('projector.views.project_milestone',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/milestones/$',
        view='MilestoneListView',
        name='projector_project_milestones'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/milestones/create/$',
        view='MilestoneCreateView',
        name='projector_project_milestone_add'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/milestones/(?P<milestone_slug>[-\w]+)/$',
        view='MilestoneDetailView',
        name='projector_project_milestone_detail'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/milestones/(?P<milestone_slug>[-\w]+)/edit/$',
        view='MilestoneEditView',
        name='projector_project_milestone_edit'),
)

# Components
urlpatterns += patterns('projector.views.project_component',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/components/$',
        view='ComponentListView',
        name='projector_project_components'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/components/create/$',
        view='ComponentCreateView',
        name='projector_project_component_add'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/components/(?P<component_slug>[-\w]+)/$',
        view='ComponentDetailView',
        name='projector_project_component_detail'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/components/(?P<component_slug>[-\w]+)/edit/$',
        view='ComponentEditView',
        name='projector_project_component_edit'),
)

# Workflow
urlpatterns += patterns('projector.views.project_workflow',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/workflow/$',
        view='WorkflowDetailView',
        name='projector_project_workflow_detail'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/workflow/edit/$',
        view='WorkflowEditView',
        name='projector_project_workflow_edit'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/workflow/create-status/$',
        view='WorkflowAddStatusView',
        name='projector_project_workflow_add_status'),
)

# Members
urlpatterns += patterns('projector.views.project_member',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/members/$',
        view='MemberListView',
        name='projector_project_member'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/members/add/$',
        view='MemberAddView',
        name='projector_project_member_add'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/members/(?P<member_username>\w+)/$',
        view='MemberEditView',
        name='projector_project_member_edit'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/members/(?P<member_username>\w+)/delete/$',
        view='MemberDeleteView',
        name='projector_project_member_delete'),
)

# Teams
urlpatterns += patterns('projector.views.project_team',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/teams/$',
        view='TeamListView',
        name='projector_project_teams'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/teams/add/$',
        view='TeamAddView',
        name='projector_project_teams_add'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/teams/(?P<name>[-_ \w]+)/$',
        view='TeamEditView',
        name='projector_project_teams_edit'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/teams/(?P<name>[-_ \w]+)/delete/$',
        view='TeamDeleteView',
        name='projector_project_teams_delete'),
)

# Project's repository browsing
urlpatterns += patterns('projector.views.project_repository',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/diff/(?P<revision_old>[\w]*)-(?P<revision_new>[\w]*)/(?P<rel_repo_url>.*)$',
        view='RepositoryFileDiff',
        name='projector_project_sources_diff'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/raw/(?P<revision>[\w]*)/(?P<rel_repo_url>.*)$',
        view='RepositoryFileRaw',
        name='projector_project_sources_raw'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/annotate/(?P<revision>[\w]*)/(?P<rel_repo_url>.*)$',
        view='RepositoryFileAnnotate',
        name='projector_project_sources_annotate'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/changesets/$',
        view='RepositoryChangesetList',
        name='projector_project_changesets'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/changesets/(?P<revision>[\w]*)/$',
        view='RepositoryChangesetDetail',
        name='projector_project_changeset_detail'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/(?P<revision>[\w]*)/(?P<rel_repo_url>.*)$',
        view='RepositoryBrowse',
        name='projector_project_sources_browse'),

    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/src/$',
        view='RepositoryBrowse',
        name='projector_project_sources'),
)

# Tasks
urlpatterns += patterns('projector.views.project_task',
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/$',
        view='TaskListView',
        name='projector_task_list'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/data/$',
        view='TaskListDataView',
        name='projector_task_list_data'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/create/$',
        view='TaskCreateView',
        name='projector_task_create'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/(?P<task_id>\d+)/$',
        view='TaskDetailView',
        name='projector_task_detail'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/(?P<task_id>\d+)/edit/$',
        view='TaskEditView',
        name='projector_task_edit'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/(?P<task_id>\d+)/watch/$',
        view='TaskWatchView',
        name='projector_task_watch'),
    url(r'^(?P<username>[-\w]+)/(?P<project_slug>[-\w]+)/tasks/(?P<task_id>\d+)/unwatch/$',
        view='TaskUnwatchView',
        name='projector_task_unwatch'),
)

urlpatterns += patterns('projector.views.users',
    url(r'^(?P<username>\w+)/$',
        view = 'UserProfileDetailView',
        name = 'projector_users_profile_detail'),
)
# ========== #
# Feeds dict #
# ========== #

feeds = {
    'projects': LatestProjectsFeed,
}

urlpatterns += patterns('',
    (r'^feeds/(?P<url>.*)/$', 'django.contrib.syndication.views.feed',
        {'feed_dict': feeds}),
)




from django import forms
from django.utils.translation import ugettext as _

from projector.core.exceptions import ProjectorError

class BaseExternalForkForm(forms.Form):
    """
    Base form class for all forms used to create *external* fork.
    """

    help_text = ''

    as_private = forms.BooleanField(initial=False, label=_('As private'),
        help_text=_('Forks project as your private project'), required=False)

    def is_public(self):
        if not hasattr(self, 'cleaned_data'):
            raise ProjectorError("Must be called after form is validated")
        return not self.cleaned_data['as_private']

    def is_private(self):
        return not self.is_public()

    def get_url(self):
        """
        Should return url to be set as ``fork_url`` attribute of the
        ``Project`` model.
        """
        raise NotImplementedError("Must be implemented in order to let project "
            "instance know where it was forked from")

    def fork(self):
        """
        Actions needed to fork the project. Real fork command should be fired by
        ``Project``'s ``post_save`` signal handler.

        This method should raise only subclass of type:
        ``projector.core.exceptions.ProjectorError``.  That way all errors
        would be returned to the user if he/she would try to fork project from
        external location and process fails. It's not validation error - form
        may validate properly. This errors should be raised while trying to
        fork from external location. We can get various errors here, such as
        problem with network etc..

        """
        raise NotImplementedError("ForkForm requires ``fork`` method to "
            "implements logic required for the forking process")




import os
import logging

from django.conf import settings

abspath = lambda *p: os.path.abspath(os.path.join(*p))

ALWAYS_SEND_MAILS_TO_MEMBERS = getattr(settings,
    'PROJECTOR_ALWAYS_SEND_MAILS_TO_MEMBERS', False)

BANNED_PROJECT_NAMES = getattr(settings, 'PROJECTOR_BANNED_PROJECT_NAMES', ())
BANNED_PROJECT_NAMES += (
    'account', 'accounts',
    'add',
    'admin', 'admins',
    'api',
    'author', 'authors',
    'ban',
    'category', 'categories',
    'change',
    'community',
    'create',
    'default',
    'delete',
    'edit', 'edits',
    'etc',
    'issue', 'issues',
    'job', 'jobs',
    'mail', 'mails',
    'message', 'messages',
    'manager', 'managers',
    'private',
    'profile', 'profiles',
    'projects',
    'register', 'registration',
    'remove',
    'task', 'tasks',
    'update',
    'user', 'users',
    'view',
)

BASIC_AUTH_REALM = getattr(settings,
    'PROJECTOR_BASIC_AUTH_REALM', 'Projector Basic Auth')

CHANGESETS_PAGINATE_BY = getattr(settings,
    'PROJECTOR_CHANGESETS_PAGINATE_BY', 10)

CREATE_PROJECT_ASYNCHRONOUSLY = getattr(settings,
    'PROJECTOR_CREATE_PROJECT_ASYNCHRONOUSLY', True)

CREATE_REPOSITORIES = getattr(settings, 'PROJECTOR_CREATE_REPOSITORIES', True)

DEFAULT_PROJECT_WORKFLOW = getattr(settings,
    'PROJECTOR_DEFAULT_PROJECT_WORKFLOW',
    'projector.conf.workflow.DefaultWorkflow')

DEFAULT_VCS_BACKEND = getattr(settings, 'PROJECTOR_DEFAULT_VCS_BACKEND', 'hg')

EDITABLE_PERMISSIONS = getattr(settings,
    'PROJECTOR_EDITABLE_PERMISSIONS',
    (
        'change_project',
        'admin_project',
        'view_project',
        'can_read_repository',
        'can_write_to_repository',
        'can_change_description',
        'can_change_category',
        'can_add_task',
        'can_change_task',
        'can_delete_task',
        'can_view_tasks',
        'can_add_member',
        'can_change_member',
        'can_delete_member',
        'can_add_team',
        'can_change_team',
        'can_delete_team',
    )
)

ENABLED_VCS_BACKENDS = getattr(settings,
    'PROJECTOR_ENABLED_VCS_BACKENDS', ['hg'])

FROM_EMAIL_ADDRESS = settings.DEFAULT_FROM_EMAIL

FORK_EXTERNAL_ENABLED = getattr(settings,
    'PROJECTOR_FORK_EXTERNAL_ENABLED', False)

FORK_EXTERNAL_MAP = getattr(settings,
    'PROJECTOR_FORK_EXTERNAL_MAP',
    {
        'bitbucket.org': 'projector.forks.bitbucket.BitbucketForkForm',
    }
)

HIDDEN_EMAIL_SUBSTITUTION = getattr(settings,
    'PROJECTOR_HIDDEN_EMAIL_SUBSTITUTION', u'email')

MAX_PROJECTS_PER_USER = getattr(settings,
    'PROJECTOR_MAX_PROJECTS_PER_USER', 50)

MILESTONE_DEADLINE_DELTA = getattr(settings,
    'PROJECTOR_MILESTONE_DEADLINE_DELTA', 60)

MILIS_BETWEEN_PROJECT_CREATION = getattr(settings,
    'PROJECTOR_MILIS_BETWEEN_PROJECT_CREATION', 15*1000)

PROJECTS_ROOT_DIR = getattr(settings, 'PROJECTOR_PROJECTS_ROOT_DIR', None)
if PROJECTS_ROOT_DIR is None:
    logging.debug("django-projector: PROJECTOR_PROJECTS_ROOT_DIR not set "
        "- will not create repositories")
else:
    PROJECTS_ROOT_DIR = abspath(PROJECTS_ROOT_DIR)
    if not os.path.exists(PROJECTS_ROOT_DIR):
        os.mkdir(PROJECTS_ROOT_DIR)
        logging.info("django-projector: created %s directory"
            % PROJECTS_ROOT_DIR)

PROJECTS_HOMEDIR_GETTER = getattr(settings, 'PROJECTOR_PROJECTS_HOMEDIR_GETTER',
    'projector.utils.helpers.get_homedir')

PRIVATE_ONLY = getattr(settings,
    'PROJECTOR_PRIVATE_ONLY', False)

SEND_MAIL_ASYNCHRONOUSELY = getattr(settings,
    'PROJECTOR_SEND_MAIL_ASYNCHRONOUSELY', True)

TASK_EMAIL_SUBJECT_SUMMARY_FORMAT = getattr(settings,
    'PROJECTOR_TASK_EMAIL_SUBJECT_SUMMARY_FORMAT',
    "[$project] #$id: $summary")

# =================== #
# Settings dictionary #
# =================== #

PROJECTOR = {
    'ALWAYS_SEND_MAILS_TO_MEMBERS': 'ALWAYS_SEND_MAILS_TO_MEMBERS',
    'BANNED_PROJECT_NAMES': BANNED_PROJECT_NAMES,
    'BASIC_AUTH_REALM': BASIC_AUTH_REALM,
    'CHANGESETS_PAGINATE_BY': CHANGESETS_PAGINATE_BY,
    'CREATE_PROJECT_ASYNCHRONOUSLY': CREATE_PROJECT_ASYNCHRONOUSLY,
    'CREATE_REPOSITORIES': CREATE_REPOSITORIES,
    'EDITABLE_PERMISSIONS': EDITABLE_PERMISSIONS,
    'ENABLED_VCS_BACKENDS': ENABLED_VCS_BACKENDS,
    'FORK_EXTERNAL_ENABLED': FORK_EXTERNAL_ENABLED,
    'FORK_EXTERNAL_MAP': FORK_EXTERNAL_MAP,
    'FROM_EMAIL_ADDRESS': settings.DEFAULT_FROM_EMAIL,
    'MAX_PROJECTS_PER_USER': MAX_PROJECTS_PER_USER,
    'MILESTONE_DEADLINE_DELTA': MILESTONE_DEADLINE_DELTA,
    'MILIS_BETWEEN_PROJECT_CREATION': MILIS_BETWEEN_PROJECT_CREATION,
    'PRIVATE_ONLY': PRIVATE_ONLY,
    'PROJECTS_ROOT_DIR': PROJECTS_ROOT_DIR,
    'PROJECTS_HOMEDIR_GETTER': PROJECTS_HOMEDIR_GETTER,
    'TASK_EMAIL_SUBJECT_SUMMARY_FORMAT': TASK_EMAIL_SUBJECT_SUMMARY_FORMAT,
}

def get_config_value(key):
    """
    Returns value of the setting pointed by the given parameter. ``key`` may
    be given without common prefix (``PROJECTOR_``). Using this method is a
    preferred way to retrieve ``projector``'s configuration values.
    """
    if key.startswith('PROJECTOR_'):
        key = key[key.find('PROJECTOR_'):]
    return globals()[key]




from django.shortcuts import redirect
from django.utils.translation import ugettext as _
from django.contrib import messages
from django.db.models import Q

from projector.models import Status, Transition
from projector.views.project import ProjectView
from projector.forms import StatusForm, StatusFormSet

from richtemplates.shortcuts import get_first_or_None

class WorkflowDetailView(ProjectView):
    """
    Returns project's workflow detail view.
    """

    template_name = 'projector/project/workflow/detail.html'

    def response(self, request, username, project_slug):
        context = {
            'project': self.project,
            # indicates that this is workflow detail page at templates
            'workflow': True,
        }
        return context

class WorkflowEditView(ProjectView):
    """
    Edits chosen project's workflow.
    """

    template_name = 'projector/project/workflow/edit.html'
    perms = ['view_project', 'change_project']

    def response(self, request, username, project_slug):
        formset = StatusFormSet(request.POST or None,
            queryset=Status.objects.filter(project=self.project))
        if request.method == 'POST':
            if formset.is_valid():
                msg = _("Workflow updated successfully")
                messages.success(request, msg)
                for form in formset.forms:
                    # update status instance
                    form.instance.save()
                    destinations = form.cleaned_data['destinations']
                    # remove unchecked
                    Transition.objects.filter(~Q(destination__in=destinations),
                        source=form.instance)\
                        .delete()
                    # add new
                    for destination in destinations:
                        Transition.objects.get_or_create(source=form.instance,
                            destination=destination)
            else:
                msg = _("Errors occured while processing formset")
                messages.error(request, msg)
                for error in formset.non_form_errors():
                    messages.error(request, error)
        context = {
            'project': self.project,
            'formset': formset,
        }
        return context

class WorkflowAddStatusView(ProjectView):
    """
    Adds status for project's workflow.
    """

    template_name = 'projector/project/workflow/add_status.html'
    perms = ['view_project']

    def response(self, request, username, project_slug):
        _max_order_status = get_first_or_None(
            self.project.status_set.only('order').order_by('-order'))
        status = Status(project=self.project,
            order=_max_order_status and _max_order_status.order+1 or 1)
        form = StatusForm(request.POST or None, instance=status)
        if request.method == 'POST' and form.is_valid():
            form.save()
            msg = _("Status added successfully")
            messages.success(request, msg)
            return redirect(self.project.get_workflow_url())
        context = {
            'project': self.project,
            'form': form,
        }
        return context




from django.contrib.auth.models import User, Group, AnonymousUser
from django.core.exceptions import ValidationError

from projector.tests.base import ProjectorTestCase
from projector.models import Team, Project


class UserToTeamConversion(ProjectorTestCase):

    def setUp(self):
        self.jack = User.objects.create_user(
            username = 'jack',
            email = 'jack@example.com',
            password = 'jack')

    def test_conversion(self):
        profile = self.jack.get_profile()
        self.assertFalse(profile.is_team)
        self.assertEqual(profile.group, None)

        project1 = Project.objects.create(name='project1', author=self.jack)
        project2 = Project.objects.create(name='project2', author=self.jack)
        group = Team.objects.convert_from_user(self.jack)

        self.assertTrue(profile.is_team)
        self.assertEqual(profile.group, group)
        self.assertTrue(group in self.jack.groups.all())

        self.assertTrue(group in Group.objects.filter(team__project=project1))
        self.assertTrue(group in Group.objects.filter(team__project=project2))

    def test_integrity(self):
        Group.objects.create(name='jack')
        try:
            Team.objects.convert_from_user(self.jack)
        except ValidationError:
            pass
        else:
            self.fail("Cannot allow to convert user into team if there is "
                      "a group with same name already persisted")

    def test_anonymous(self):
        user = AnonymousUser()
        try:
            Team.objects.convert_from_user(user)
        except ValidationError:
            pass
        else:
            self.fail("Cannot allow to convert anonymous user to team")

    def test_staff(self):
        self.jack.is_staff = True
        self.jack.save()
        try:
            Team.objects.convert_from_user(self.jack)
        except ValidationError:
            pass
        else:
            self.fail("Cannot allow to convert staff member to team")

    def test_superuser(self):
        self.jack.is_superuser = True
        self.jack.save()
        try:
            Team.objects.convert_from_user(self.jack)
        except ValidationError:
            pass
        else:
            self.fail("Cannot allow to convert superuser to team")

    def test_non_active(self):
        self.jack.is_active = False
        self.jack.save()
        try:
            Team.objects.convert_from_user(self.jack)
        except ValidationError:
            pass
        else:
            self.fail("Cannot allow to convert inactive user to team")





"""
Management commands listeners needs to be stored within ``management`` module
inside app.
"""
import sys
import logging

from django.db.models import Count

from projector.models import Project, Config
from projector.settings import get_config_value

from guardian.shortcuts import get_perms, get_perms_for_model, assign

def update_project_permissions(sender, **kwargs):
    """
    Creates missing permissions for projects' authors.
    """
    project_permissions = get_perms_for_model(Project)
    project_perms_set = set((p.codename for p in project_permissions))
    for project in Project.objects.all().select_related('author'):
        perms = get_perms(project.author, project)
        perms_set = set(perms)

        missing_perms = project_perms_set.difference(perms_set)
        missing_perms.discard('add_project') # This is not project specific
        for perm in missing_perms:
            new_perm = assign(perm, project.author, project)
            msg = '[INFO] Adding permission %s' % new_perm
            if kwargs['verbosity'] >= 2:
                print msg

def put_missing_project_configs(sender, **kwargs):
    """
    Required for backword compatibility as Config is a new model.
    """
    project_count = Project.objects.count()
    config_count = Config.objects.count()

    if project_count > config_count:
        if kwargs['interactive'] is True:
            msg = ("There are %d projects without configuration"
                    % (project_count - config_count))
            logging.info(msg)
            answer = ''
            while answer.lower() not in ('yes', 'no', 'y', 'n'):
                prompt = 'Create missing configs [yes/no, default=yes]: '
                try:
                    answer = raw_input(prompt).lower()
                except (KeyboardInterrupt, EOFError):
                    sys.stderr.write('\nInterrupted by user - taken as "no"\n')
                    answer = 'no'
                if answer == '':
                    answer = 'yes'
                if answer in ('y', 'yes'):
                    projects = Project.objects\
                        .annotate(config_count=Count('config'))\
                        .filter(config_count=0)
                    for project in projects:
                        config = project.create_config()
                        msg = "[INFO] Created %s" % config
                        if kwargs['verbosity'] >= 1:
                            print msg
                elif answer in ('n', 'no'):
                    sys.stderr.write("[WARNING] Projector app may couse errors "
                        "due to missing config for projects\n")

def create_missing_repositories(sender, **kwargs):
    """
    If :setting:`CREATE_REPOSITORIES` is set to ``True`` and we found
    :model:`Project` instances without repository we need to create them.
    """
    projects = Project.objects.filter(repository=None)

    if projects.count() > 0:
        if kwargs['interactive'] is True:
            msg = "There are %d projects without repos" % projects.count()
            logging.info(msg)
            answer = ''
            while answer.lower() not in ('yes', 'y', 'no', 'n'):
                prompt = ('Create missing repos [yes/no, default=yes] '
                          '[default vcs backend is: %s]'
                          % get_config_value('DEFAULT_VCS_BACKEND'))
                try:
                    answer = raw_input(prompt).lower()
                except (KeyboardInterrupt, EOFError):
                    sys.stderr.write('\nInterrupted by user - taken as "no"\n')
                    answer = 'no'

                if answer == '':
                    answer = 'yes'

                if answer in ('y', 'yes'):
                    for project in projects:
                        repo = project.create_repository()
                        msg = "[INFO] Created %s" % repo
                        if kwargs['verbosity'] >= 1:
                            print msg
                elif answer in ('n', 'no'):
                    sys.stderr.write("Answered 'no'. There are still %d missing"
                                     " repositories" % projects.count())




import os
import sys

from django.conf import global_settings

abspath = lambda *p: os.path.abspath(os.path.join(*p))

DEBUG = True

PROJECT_ROOT = abspath(os.path.dirname(__file__))
PROJECTOR_MODULE_PATH = abspath(PROJECT_ROOT, '..')
sys.path.insert(0, PROJECTOR_MODULE_PATH)

TEST_RUNNER = 'django.test.simple.DjangoTestSuiteRunner'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': abspath(PROJECT_ROOT, '.hidden.db'),
        'TEST_NAME': ':memory:',
    },
}

# Make sqlite3 files relative to project's directory
for db, conf in DATABASES.items():
    if conf['ENGINE'] == 'sqlite3' and not conf['NAME'].startswith(':'):
        conf['NAME'] = abspath(PROJECT_ROOT, conf['NAME'])

INSTALLED_APPS = (
    'admin_tools',
    'admin_tools.theming',
    'admin_tools.menu',
    'admin_tools.dashboard',

    'native_tags',

    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.comments',
    'django.contrib.markup',
    'django.contrib.messages',
    'django.contrib.webdesign',


    # External
    'djalog',
    'django_extensions',
    'django_sorting',
    'djcelery',
    'djcelery_email',
    'ghettoq',
    'gravatar',
    'guardian',
    'pagination',
    'registration',
    'richtemplates',
    'projector',
    'vcs.web.simplevcs',

    'example_project',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.transaction.TransactionMiddleware',

    'djalog.middleware.SQLLoggingMiddleware',
    'richtemplates.middleware.Http403Middleware',
    'django_sorting.middleware.SortingMiddleware',
    'vcs.web.simplevcs.middleware.PaginationMiddleware',
)

INTERNAL_IPS = ('127.0.0.1',)

MEDIA_ROOT = abspath(PROJECT_ROOT, 'media')
MEDIA_URL = '/media/'
ADMIN_MEDIA_PREFIX = '/admin-media/'

ROOT_URLCONF = 'example_project.urls'

TEMPLATE_CONTEXT_PROCESSORS = global_settings.TEMPLATE_CONTEXT_PROCESSORS + (
    'django.core.context_processors.request',
    'richtemplates.context_processors.media',
)
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.load_template_source',
    'django.template.loaders.app_directories.load_template_source',
    'django.template.loaders.eggs.load_template_source',
)

TEMPLATE_DIRS = (
    os.path.join(os.path.dirname(__file__), 'templates'),
)

SITE_ID = 1

USE_I18N = True
USE_L10N = True

CACHE_PREFIX = 'projector-example-project'
#CACHE_TIMEOUT = 1 # For dev server

LOGIN_REDIRECT_URL = '/'
AUTH_PROFILE_MODULE = 'projector.UserProfile'

# ================== #
# PROJECTOR SETTINGS #
# ================== #

PROJECTOR_PROJECTS_ROOT_DIR = abspath(
    PROJECT_ROOT, 'projects')
PROJECTOR_BANNED_PROJECT_NAMES = ('barfoo',)
PROJECTOR_SEND_MAIL_ASYNCHRONOUSELY = True
PROJECTOR_CREATE_PROJECT_ASYNCHRONOUSLY = True

# =============== #
# DJALOG SETTINGS #
# =============== #

DJALOG_SQL = True
DJALOG_SQL_SUMMARY_ONLY = True
DJALOG_LEVEL = 5
DJALOG_USE_COLORS = True
DJALOG_FORMAT = "[%(levelname)s] %(message)s"

# ====================== #
# RICHTEMPLATES SETTINGS #
# ====================== #

RICHTEMPLATES_RESTRUCTUREDTEXT_DIRECTIVES = {
    'code-block': 'richtemplates.rstdirectives.pygments_directive',
}
RICHTEMPLATES_DEFAULT_SKIN = 'ruby'
RICHTEMPLATES_PYGMENTS_STYLES = {
    'irblack': 'richtemplates.pygstyles.irblack.IrBlackStyle',
}

# ==================== #
# NATIVE_TAGS SETTINGS #
# ==================== #

NATIVE_TAGS = (
    'richtemplates.templatetags.native',
    'projector.templatetags.native',
)

# ====================== #
# DEBUG TOOLBAR SETTINGS #
# ====================== #

DEBUG_TOOLBAR_CONFIG = {
    'INTERCEPT_REDIRECTS': False,
}

# ================ #
# REQUEST SETTINGS #
# ================ #

REQUEST_IGNORE_PATHS = (
    r'^%s' % MEDIA_URL.lstrip('/'),
    r'^%s' % ADMIN_MEDIA_PREFIX.lstrip('/'),
)

# ============== #
# EMAIL SETTINGS #
# ============== #

DEFAULT_FROM_EMAIL = 'webmaster@localhost'

EMAIL_HOST = 'localhost'
EMAIL_HOST_PASSWORD = ''
EMAIL_HOST_USER = ''
EMAIL_PORT = 25
EMAIL_SUBJECT_PREFIX = '[Django] '
EMAIL_USE_TLS = False

EMAIL_BACKEND = 'djcelery_email.backends.CeleryEmailBackend'

CELERY_EMAIL_BACKEND = 'django.core.mail.backends.locmem.EmailBackend'
CELERY_EMAIL_TASK_CONFIG = {
    'queue' : 'email',
    'rate_limit' : '60/m', # 60 emails per minute
}

# ======================= #
# AUTHENTICATION SETTINGS #
# ======================= #

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend', # this is default
    'guardian.backends.ObjectPermissionBackend',
)
ANONYMOUS_USER_ID = -1

ACCOUNT_ACTIVATION_DAYS = 7
GRAVATAR_DEFAULT_IMAGE = 'mm'

try:
    from conf.local_settings import *
    try:
        for app in LOCAL_INSTALLED_APPS:
            if app not in INSTALLED_APPS:
                INSTALLED_APPS += (app,)
        for middleware in LOCAL_MIDDLEWARE_CLASSES:
            if middleware not in MIDDLEWARE_CLASSES:
                MIDDLEWARE_CLASSES += (middleware,)
    except NameError:
        pass
except ImportError:
    pass


# ================ #
# CELLERY SETTINGS #
# ================ #

CARROT_BACKEND = "ghettoq.taproot.Database"

BROKER_CONNECTION_MAX_RETRIES = 0

CELERY_ALWAYS_EAGER = False
CELERYD_MAX_TASKS_PER_CHILD = 100
CELERYD_LOG_LEVEL = 'DEBUG'




"""
Project management Django application with task
tracker and repository backend integration.
"""

VERSION = (0, 1, 8)

__version__ = '.'.join((str(each) for each in VERSION[:4]))

def get_version():
    """
    Returns shorter version (digit parts only) as string.
    """
    return '.'.join((str(each) for each in VERSION[:3]))




import os
import sys
import string
import logging
import datetime
import traceback

from decimal import Decimal

from django.conf import settings
from django.contrib.auth.models import User, Group, AnonymousUser, Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.contrib.sites.models import Site
from django.core.exceptions import ValidationError, PermissionDenied,\
    ImproperlyConfigured
from django.core.mail import send_mail
from django.db import models
from django.db.models import Q
from django.db.models.query import QuerySet
from django.template.defaultfilters import slugify
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _
from django.utils.datastructures import SortedDict

from guardian.shortcuts import assign, get_perms, get_perms_for_model

from autoslug import AutoSlugField

from projector import settings as projector_settings
from projector.core.exceptions import ProjectorError
from projector.core.exceptions import ConfigAlreadyExist
from projector.core.exceptions import ForkError
from projector.managers import ProjectManager
from projector.managers import TaskManager
from projector.managers import TeamManager
from projector.managers import WatchedItemManager
from projector.settings import get_config_value
from projector.signals import post_fork
from projector.utils import abspath, str2obj, using_projector_profile
from projector.utils.lazy import LazyProperty
from projector.utils.helpers import Choices

from vcs.backends import get_supported_backends
from vcs.exceptions import VCSError
from vcs.web.simplevcs.models import Repository

from richtemplates.models import UserProfile as RichUserProfile

from treebeard.al_tree import AL_Node


PROJECT_VCS_ALIAS_FIELD = '_vcs_alias'
PROJECT_WORKFLOW_FIELD = '_workflow_obj'

class WatchedItem(models.Model):
    """
    Projects, Tasks and other items watched by the User.
    """
    user = models.ForeignKey(User)
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    objects = WatchedItemManager()

    class Meta:
        unique_together = ('user', 'content_type', 'object_id')

    def __unicode__(self):
        return u'%s watched by %s' % (self.content_object, self.user)

class Watchable(object):
    def watch(self, user):
        item, created = WatchedItem.objects.get_or_create(
            user = user,
            content_type = ContentType.objects.get_for_model(self),
            object_id = self.pk,
        )
        return item, created

    def unwatch(self, user):
        WatchedItem.objects.filter(
            user = user,
            content_type = ContentType.objects.get_for_model(self),
            object_id = self.pk,
        ).delete()

    def is_watched(self, user):
        if isinstance(user, AnonymousUser):
            return False
        try:
            WatchedItem.objects.get(
                user = user,
                content_type = ContentType.objects.get_for_model(self),
                object_id = self.pk)
            return True
        except WatchedItem.DoesNotExist:
            return False

    def get_watchers(self):
        users = User.objects.filter(
            watcheditem__content_type = ContentType.objects.get_for_model(self),
            watcheditem__object_id = self.pk,
        )
        return users

class DictModel(models.Model):
    name = models.CharField(_('name'), max_length=32)
    description = models.TextField(_('description'), null=True, blank=True)

    def __unicode__(self):
        return self.name

    class Meta:
        abstract = True
        ordering = ('id',)

class OrderedDictModel(DictModel):
    """
    DictModel with order column and default ordering.
    """
    order = models.IntegerField(_('order'))

    class Meta:
        abstract = True
        ordering = ['order', 'name']

class ProjectCategory(models.Model):
    name = models.CharField(_('name'), max_length=64, unique=True)
    slug = models.SlugField(unique=True, editable=False, max_length=64)
    public = models.BooleanField(_('public'), default=True)

    class Meta:
        ordering = ['name']
        verbose_name = _('project')
        verbose_name_plural = _('project categories')

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.pk is None:
            self.slug = slugify(self.name)
        super(ProjectCategory, self).save(*args, **kwargs)

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_category_detail', (), {
            'username': self.project.author.username,
            'project_category_slug' : self.slug,
        })

def validate_project_name(name):
    """
    Checks if this project name may be used.
    """
    if name.strip().lower() in projector_settings.BANNED_PROJECT_NAMES:
        raise ValidationError(_("This name is restricted"))


class State(Choices):
    """
    Represents state of the project.
    """
    ERROR = -1
    PENDING = 0
    CREATED = 10
    MEMBERSHIPS_CREATED = 20
    AUTHOR_PERMISSIONS_CREATED = 30
    WORKFLOW_CREATED = 40
    CONFIG_CREATED = 50
    REPOSITORY_CREATED = 60
    READY = 100


class Project(AL_Node, Watchable):
    """
    Most important model within whole application. It provides connection with
    all other models.
    """

    name = models.CharField(_('name'), max_length=64,
        validators=[validate_project_name])
    slug = AutoSlugField(max_length=128, populate_from='name',
        unique_with='author', validators=[validate_project_name])
    category = models.ForeignKey(ProjectCategory, verbose_name=_('category'),
        null=True, blank=True)
    description = models.TextField(_('description'), null=True, blank=True)
    home_page_url = models.URLField(_("home page url "), null=True, blank=True,
        verify_exists=False)
    is_active = models.BooleanField(_('is active'), default=True)
    public = models.BooleanField(_('public'), default=True)
    members = models.ManyToManyField(User, verbose_name=_('members'),
        through="Membership")
    teams = models.ManyToManyField(Group, verbose_name=_('teams'),
        through="Team", null=True, blank=True)
    author = models.ForeignKey(User, name=_('author'),
        related_name='created_projects')
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)
    modified_at = models.DateTimeField(_('modified at'), auto_now=True)
    outdated = models.BooleanField(_('outdated'), default=False)
    repository = models.ForeignKey(Repository, null=True, blank=True,
        verbose_name=_("repository"), default=None)
    state = models.IntegerField(_('state'), choices=State.as_choices(),
        default=0)
    error_text = models.CharField(_('error text'), max_length=256, null=True,
        blank=True)

    parent = models.ForeignKey('self', related_name='children_set',
       null=True, blank=True, db_index=True)
    fork_url = models.URLField(verify_exists=False, null=True, blank=True)

    node_order_by = ['author', 'name']

    objects = ProjectManager()

    class Meta:
        verbose_name = _('project')
        verbose_name_plural = _('projects')
        ordering = ['name']
        get_latest_by = 'created_at'
        unique_together = ['author', 'name']
        permissions = (
            ('view_project', 'Can view project'),
            ('admin_project', 'Can administer project'),
            ('can_read_repository', 'Can read repository'),
            ('can_write_to_repository', 'Can write to repository'),
            ('can_change_description', 'Can change description'),
            ('can_change_category', 'Can change category'),
            ('can_add_task', 'Can add task'),
            ('can_change_task', 'Can change task'),
            ('can_delete_task', 'Can delete task'),
            ('can_view_tasks', 'Can view tasks'),
            ('can_add_member', 'Can add member'),
            ('can_change_member', 'Can change member'),
            ('can_delete_member', 'Can delete member'),
            ('can_add_team', 'Can add team'),
            ('can_change_team', 'Can change team'),
            ('can_delete_team', 'Can delete team'),
        )

    def __unicode__(self):
        return self.name

    def is_public(self):
        return self.public

    def is_private(self):
        return not self.public

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        self.full_clean()
        project = super(Project, self).save(*args, **kwargs)
        # Add necessary permissions for author - we need to do this
        # *NOT* asynchronousely as instant redirect after project creation
        # may cause it's author not to be able to see project page
        return project

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_detail', (), {
            'username': self.author.username,
            'project_slug' : self.slug,
        })

    @models.permalink
    def get_edit_url(self):
        return ('projector_project_edit', (), {
            'username': self.author.username,
            'project_slug' : self.slug,
        })

    @models.permalink
    def get_state_url(self):
        return ('projector_project_state', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_fork_url(self):
        return ('projector_project_fork', (), {
            'username': self.author.username,
            'project_slug' : self.slug,
        })

    @models.permalink
    def get_members_url(self):
        return ('projector_project_member', (), {
            'username': self.author.username,
            'project_slug' : self.slug,
        })

    @models.permalink
    def get_members_add_url(self):
        return ('projector_project_member_add', (), {
            'username': self.author.username,
            'project_slug' : self.slug,
        })

    @models.permalink
    def get_members_edit_url(self, username):
        return ('projector_project_member_edit', (), {
            'username': self.author.username,
            'project_slug': self.slug,
            'member_username': username,
        })

    @models.permalink
    def get_teams_url(self):
        return ('projector_project_teams', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_teams_add_url(self):
        return ('projector_project_teams_add', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_teams_edit_url(self, name):
        return ('projector_project_teams_edit', (), {
            'username': self.author.username,
            'project_slug': self.slug,
            'name': name,
        })

    @models.permalink
    def get_create_task_url(self):
        return ('projector_task_create', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_task_list_url(self):
        return ('projector_task_list', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_milestones_url(self):
        return ('projector_project_milestones', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_milestone_add_url(self):
        return ('projector_project_milestone_add', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_components_url(self):
        return ('projector_project_components', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_component_add_url(self):
        return ('projector_project_component_add', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_workflow_url(self):
        return ('projector_project_workflow_detail', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_workflow_edit_url(self):
        return ('projector_project_workflow_edit', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_workflow_add_status_url(self):
        return ('projector_project_workflow_add_status', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_browse_repo_url(self):
        return ('projector_project_sources', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    @models.permalink
    def get_changesets_url(self):
        return ('projector_project_changesets', (), {
            'username': self.author.username,
            'project_slug': self.slug,
        })

    def is_pending(self):
        return self.state != State.ERROR and self.state < State.READY

    def get_task(self, id):
        queryset = Task.objects.filter(project=self)
        if not queryset:
            # queryset at this point may be evaluated to empty list
            raise Task.DoesNotExist
        return queryset.get(id=id)

    def get_tasks(self):
        return self.task_set\
            .select_related('status', 'milestone', 'owner',
                    'priority', 'component')

    def get_open_tasks(self):
        return self.get_tasks().filter(status__is_resolved=False)

    def get_closed_tasks(self):
        return self.get_tasks().filter(status__is_resolved=True)

    def _get_homedir(self):
        """
        Returns directory containing all files related to this project. If
        repository is already set at this project, it's parent directory would
        be returned. Otherwise path is computed.

        This is semi *private* method as we cannot allow this to be called at
        templates (methods starting with underscore cannot be called within
        django templates).
        """
        if self.repository:
            return abspath(self.repository.path, '..')
        try:
            getter_path = get_config_value('PROJECTS_HOMEDIR_GETTER')
            getter = str2obj(getter_path)
        except ImportError, err:
            raise ImproperlyConfigured("PROJECTOR_PROJECTS_HOMEDIR_GETTER "
            "does not point to proper function. Error was: %s" % err)
        relative_path = getter(self)
        homedir = abspath(projector_settings.PROJECTS_ROOT_DIR, relative_path)
        if not homedir.endswith(os.path.sep):
            homedir += os.path.sep
        if os.path.exists(projector_settings.PROJECTS_ROOT_DIR) and \
            not os.path.exists(homedir):
            os.mkdir(homedir)
            logging.debug("Project '%s': Created homedir at %s"
                % (self, homedir))
        return homedir

    def _get_repo_path(self, vcs_alias=None):
        """
        Returns path to this project's repository. If repository is already set
        it's path is returned. Otherwise it is a join of *homedir* and vcs alias
        (which may be given as a parameter).
        """
        if self.repository:
            return self.repository.path
        if vcs_alias is not None:
            self.set_vcs_alias(vcs_alias)
        alias = self.get_vcs_alias()
        repo_path = abspath(self._get_homedir(), alias)
        return repo_path

    def get_repo_url(self):
        """
        Returns full url of the project (with domain based on
        django sites framework).
        """
        current_site = Site.objects.get_current()
        # FIXME: What should we do with hardcoded schemas?
        if settings.DEBUG:
            prefix = 'http://'
        else:
            prefix = 'https://'
        return ''.join((prefix, current_site.domain, self.get_absolute_url()))

    def add_timeline_entry(self, action, author):
        """
        Initially wanted to implement as models.signal handler -
        as used in django-issues. Ended up with simple wrapper
        for expending project's timeline.
        Some more logic may be added over time.
        """
        timeline_entry_info = {
            'project' : self,
            'action' : action,
            'user' : author,
        }
        obj = TimelineEntry.objects.create(**timeline_entry_info)
        logging.info("Craeted timeline entry: %s" % obj)

    def set_author_permissions(self):
        """
        Creates all permissions for project's author.

        If successful, project should change it's state to
        ``State.AUTHOR_PERMISSIONS_CREATED``. Note, that this method doesn't
        make any database related queries - state should be flushed manually.
        """
        perms = set([p.codename for p in get_perms_for_model(Project)
            if p.codename != 'add_project'])
        author_perms = set(get_perms(self.author, self))
        for perm in (perms - author_perms):
            assign(perm, self.author, self)

        # If author is Team we need to add perms to that Team too
        if self.author.get_profile().is_team:
            group = self.author.get_profile().group
            team_perms = set(get_perms(group, self))
            for perm in (perms - team_perms):
                assign(perm, self.author.get_profile().group, self)
        self.state = State.AUTHOR_PERMISSIONS_CREATED

    def set_memberships(self):
        """
        Creates :model:`Membership` for this project and it's author. If author
        is a team, we need to create :model:`Team` object too.

        If successful, project should change it's state to
        ``State.MEMBERSHIPS_CREATED``. Note, that this method doesn't make any
        database related queries - state should be flushed manually.
        """
        Membership.objects.create(project=self, member=self.author)
        if self.author.is_superuser:
            # we don't need to add permissions for superuser
            # as superusers has always all permissions
            return
        # If author is a team, we need to create :model:`Team` instance for
        # his/her group
        profile = self.author.get_profile()
        if profile.is_team:
            Team.objects.create(project=self, group=profile.group)
        logging.debug("Memberships created for project %s" % self)

    def set_workflow(self, workflow):
        setattr(self, PROJECT_WORKFLOW_FIELD, workflow)

    def get_workflow(self):
        workflow = getattr(self, PROJECT_WORKFLOW_FIELD, None)
        if workflow is None:
            workflow = str2obj(get_config_value('DEFAULT_PROJECT_WORKFLOW'))
        return workflow

    def create_workflow(self, workflow=None):
        """
        Creates default workflow for the project. We need to create objects
        required to work on issues (components, types, statuses with their
        transitions if custom workflow is used).

        Workflow is retrieved using ``get_workflow`` method on the instance but
        may be overridden by passing parameter directly.

        If successful, project should change it's state to
        ``State.WORKFLOW_CREATED``. Note, that this method doesn't make any
        database related queries - state should be flushed manually.

        :param workflow: python object defining tuples of dicts with
          information on project *metadata*; may be a string pointing to the
          object; if parameter is None, value would be retrieved using
          ``get_workflow`` instance's method
        """

        if workflow is None:
            workflow = self.get_workflow()

        for component_info in workflow.components:
            component, created = Component.objects\
                .get_or_create(project = self, **component_info)
            if created:
                logging.debug("For project '%s' new component '%s' was craeted"
                    % (component.project, component))
        for task_type_info in workflow.task_types:
            task, created = TaskType.objects\
                .get_or_create(project = self, **task_type_info)
            if created:
                logging.debug("For project '%s' new task type '%s' was craeted"
                    % (task.project, task))
        for priority_info in workflow.priorities:
            priority, created = Priority.objects\
                .get_or_create(project = self, **priority_info)
            if created:
                logging.debug("For project '%s' new priority '%s' was created."
                    % (priority.project, priority))

        for status_info in workflow.statuses:
            status, created = Status.objects\
                .get_or_create(project = self, **status_info)
            if created:
                logging.debug("For project '%s' new status '%s' was created."
                    % (status.project, status))
        # Create necessary transitions
        logging.debug("Creating transitions")
        self.create_all_transitions()

        self.state = State.WORKFLOW_CREATED

    def create_all_transitions(self):
        """
        Creates ``Transition`` objects linking all projects' statuses with
        each other providing default task *workflow* schema.
        """
        statuses = self.status_set.all()
        for status in statuses:
            for new_status in statuses:
                transition, created = Transition.objects\
                    .get_or_create(source=status, destination=new_status)
                if created:
                    logging.debug("Project '%s': created %s"
                        % (self, transition))

    def get_transitions(self):
        """
        Returns queryset of ``Transition`` objects related with this project.
        """
        transitions = Transition.objects\
            .filter(source__in=self.status_set.all())
        return transitions

    def set_vcs_alias(self, vcs_alias):
        """
        Prepares this project for repository creation process. Note that
        :error:`ProjectorError` is raised if repository for this instance has
        been already created. Same exception is raised if ``vcs`` does not support
        backend for the given alias.

        :param vcs_alias: alias for ``vcs`` backend
        """
        supported_backends = get_supported_backends()
        if vcs_alias not in supported_backends:
            raise ProjectorError("Unsupported backend %s. Supported backends "
            "have following aliases: %s" % (vcs_alias,
            ', '.join(supported_backends)))
        if self.repository is not None:
            raise ProjectorError("Cannot prepare project for repository "
            "creation if it is already set")
        setattr(self, PROJECT_VCS_ALIAS_FIELD, vcs_alias)

    def get_vcs_alias(self):
        """
        Returns vcs_alias which should be already configured for this project.
        Defaulting to :setting:`PROJECTOR_DEFAULT_VCS_BACKEND`.
        """
        if self.repository is not None:
            return self.repository.alias
        default = get_config_value('DEFAULT_VCS_BACKEND')
        vcs_alias = getattr(self, PROJECT_VCS_ALIAS_FIELD, default)
        return vcs_alias

    def create_repository(self, vcs_alias=None):
        """
        Creates repository for this project.

        ``vcs_alias`` is retrieved using ``get_vcs_alias`` method but may be
        overridden by passing parameter directly.

        If successful, project should change it's state to
        ``State.REPOSITORY_CREATED``. Note, that this method doesn't make any
        database related queries - state should be flushed manually.

        :param vcs_alias: alias for ``vcs`` backend, if None given this value
          is retrieved using ``get_vcs_alias`` instance's method

        :raise ImproperlyConfigured: if given ``vcs_alias`` is not one of
          enabled backends

        :raise ProjectorError: if repository is already created, or given
          ``vcs_alias`` is not one of vcs supported backends
        """
        if not get_config_value('CREATE_REPOSITORIES'):
            raise ProjectorError("Cannot create repository if "
            "PROJECTOR_CREATE_REPOSITORIES is not set to True")

        if self.repository is not None:
            raise ProjectorError("Cannot create repository if project has one "
            "already")

        if vcs_alias is None:
            vcs_alias = self.get_vcs_alias()

        supported_backends = get_supported_backends()
        if vcs_alias not in supported_backends:
            raise ProjectorError("Unsupported backend %s. Supported backends "
            "have following aliases: %s" % (vcs_alias,
            ', '.join(supported_backends)))

        enabled_backends = get_config_value('ENABLED_VCS_BACKENDS')
        if vcs_alias not in enabled_backends:
            raise ImproperlyConfigured("Cannot use VCS backend not specified at"
            "PROJECTOR_ENABLED_VCS_BACKENDS. Tried %s and available backends "
            "are: %s" % (vcs_alias, ', '.join(enabled_backends)))

        # Repository creation process
        try:
            clone_url = None
            if self.parent:
                # Attempt to fork internally
                clone_url = self.parent._get_repo_path()
            elif self.fork_url:
                # Attempt to fork from external location
                clone_url = self.fork_url
            repository = Repository.objects.create(
                path=self._get_repo_path(vcs_alias),
                alias=vcs_alias, clone_url=clone_url)
            # Update is much faster than save
            self.repository = repository
            Project.objects.filter(pk=self.pk).update(repository=repository)
            return repository
        except VCSError, err:
            traceback_msg = '\n'.join(traceback.format_exception(*
                (sys.exc_info())))
            msg = "Couldn't create repository"
            org_msg = msg + ". Original error was: %s" % err
            org_msg = '\n\n'.join((msg, traceback_msg))
            logging.error(org_msg)
            if settings.DEBUG:
                # Show original error only in debug mode
                raise ProjectorError(org_msg)
            else:
                raise ProjectorError(msg)
        else:
            logging.debug("%s created for project %s" % (repository, self))
            self.state = State.REPOSITORY_CREATED

    def create_config(self):
        """
        Creates default configuration for given project.

        If successful, project should change it's state to
        ``State.CONFIG_CREATED``. Note, that this method doesn't make any
        database related queries - state should be flushed manually.
        """
        if self.config_set.count() > 0:
            raise ConfigAlreadyExist("Project %s with id %d has already "
                "related configuration" % (self, self.id))
        config = Config.objects.create(
            project = self,
            editor = self.author,
        )
        self.state = State.CONFIG_CREATED
        return config

    def get_config(self):
        """
        Returns Config for this project.
        """
        return Config.objects.get(project=self)
    config = property(get_config)

    def setup(self, vcs_alias=None, workflow=None):
        """
        Should be called **AFTER** instance is saved into database as all
        methods here creates necessary models for the project and in order
        to create relations it is needed that instance is persisted first.
        """
        # Prepare if parametrs are given. Otherwise assume that preparation
        # methods have been called already
        if vcs_alias:
            self.set_vcs_alias(vcs_alias)
        if workflow:
            self.set_workflow(workflow)

        # Fire up methods
        self.set_memberships()
        self.set_author_permissions()
        self.create_workflow()
        self.create_config()

        if get_config_value('CREATE_REPOSITORIES'):
            self.create_repository(vcs_alias)

    def get_watchers(self):
        watchers = super(Project, self).get_watchers()
        watchers = watchers.filter(models.Q(
            membership__project=self) |
            models.Q(groups__team__project=self))
        return watchers

    def fork(self, user, force_private=False):
        """
        Creates fork from internal context. As we use django-treebeard,
        we need to use it's api to create a child.

        :param user: new author for forked project
        :param force_private: if set to True, project would be created
          as private

        :returns: :py:class:`projector.models.Project`

        :raise ``django.core.exceptions.PermissionDenied``: when
          user is anonymous, not active or cannot access this project
        :raise ``projector.core.exceptions.ForkError``: when
          user has already forked this project
        """
        if user.is_anonymous() or not user.is_active:
            raise PermissionDenied("Fork is allowed for active users only")
        if not self in Project.objects.for_user(user):
            raise PermissionDenied("User is not allowed to fork this project")
        if user == self.author:
            raise ForkError("Author cannot fork own project")
        user_fork = self.get_fork_for_user(user)
        if user_fork:
            raise ForkError("User already forked this project")

        if force_private:
            is_public = False
        else:
            is_public = self.public

        project_info = {
            'name': self.name,
            'author': user,
            'public': is_public,
            'fork_url': self.get_repo_url(),
        }
        forked = self.add_child(**project_info)
        post_fork.send(sender=self, fork=forked)
        return forked

    def get_fork_for_user(self, user):
        """
        Returns fork of this project's root for the given user. If user haven't
        forked project from this instance's tree, ``None`` is returned.
        """
        for fork in self.get_all_forks():
            if fork.author_id == user.id:
                return fork
        return None

    def get_all_forks(self):
        """
        Returns all forks for this project, starting from root. Returned object
        is a list (not ``Queryset``) and contains this instance.
        """
        root = self.get_root()
        forks = Project.get_tree(parent=root)
        return forks

    def is_fork(self):
        """
        Returns ``True`` if this instance is a fork (has parent), ``False``
        otherwise.
        """
        return self.parent is not None or self.fork_url

class Config(models.Model):
    """
    This model stores configuration on "per project" basis.
    """
    project = models.ForeignKey(Project, unique=True, verbose_name=_('Project'))
    edited_at = models.DateTimeField(auto_now=True, verbose_name=('Edited at'))
    editor = models.ForeignKey(User)
    basic_realm = models.CharField(max_length=64, default='Basic Auth Realm',
        verbose_name = _('Basic realm'))
    always_mail_members = models.BooleanField(default = False,
        verbose_name = _('Always send mail to members'))
    changesets_paginate_by = models.PositiveIntegerField(
        verbose_name = _('Changesets paginate by'), default = 10)
    from_email_address = models.EmailField(max_length = 64,
        default = get_config_value('FROM_EMAIL_ADDRESS'),
        verbose_name = _('Email sender address'))
    #Should be scaped with from django.utils.html.escape
    task_email_summary_format = models.CharField(max_length = 128,
        default = get_config_value('TASK_EMAIL_SUBJECT_SUMMARY_FORMAT'),
        verbose_name = ('Task email summary format'),
        help_text = _(
            'You may use following placeholders: $project $summary $id'))
    milestone_deadline_delta = models.PositiveIntegerField(
        default = get_config_value('MILESTONE_DEADLINE_DELTA'),
        verbose_name = _('Milestone deadline delta'),
        help_text = _(
            u'Every milestone has its deadline and this number specifies '
            u'how many days would be given by default. It may be set to '
            u'any date during milestone creation process though.',
        ))

    def __unicode__(self):
        return u'<Config for %s>' % self.project


class Component(models.Model):
    project = models.ForeignKey(Project)
    name = models.CharField(max_length=64)
    slug = AutoSlugField(max_length=64, populate_from='name',
        always_update=True, unique_with='project')
    description = models.TextField(null=True, blank=True)

    class Meta:
        verbose_name = _('component')
        verbose_name_plural = _('components')
        ordering = ('name',)
        unique_together = ('project', 'name')

    def __unicode__(self):
        return self.name

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_component_detail', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'component_slug': self.slug,
        })

    @models.permalink
    def get_edit_url(self):
        return ('projector_project_component_edit', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'component_slug': self.slug,
        })

class Membership(models.Model):
    member = models.ForeignKey(User, verbose_name=_('member'))
    project = models.ForeignKey(Project, verbose_name=_('project'))
    joined_at = models.DateTimeField(_('joined at'), auto_now_add=True)

    def __unicode__(self):
        return u"%s@%s" % (self.member, self.project)

    class Meta:
        unique_together = ('project', 'member')

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_member_edit', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'member_username': self.member.username,
        })

    @models.permalink
    def get_delete_url(self):
        return ('projector_project_member_delete', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'member_username': self.member.username,
        })

    @LazyProperty
    def perms(self):
        """
        Returns Permission objects for member, not his/her groups.
        """
        return get_perms(self.member, self.project)

    @LazyProperty
    def all_perms(self):
        """
        Returns all Permission objects for member's and
        his/he groups (to fetch user specific permissions only, use ``perms``
        instead).
        """
        perms = Permission.objects.filter(
            models.Q(group__in = self.member.groups.all()) |
            models.Q(user=self.member)
        ).filter(
            approved = True,
            content_type = ContentType.objects.get_for_model(Project),
            object_id = self.project_id,
        )
        return perms

    @LazyProperty
    def teams(self):
        """
        Returns all Team instances related with membership's project and member
        by his/her groups.
        """
        return Team.objects.filter(
            project=self.project,
            group__in=self.member.groups.all()
        )


class Team(models.Model):
    group = models.ForeignKey(Group, verbose_name=_('group'))
    project = models.ForeignKey(Project, verbose_name=_('project'))
    joined_at = models.DateTimeField(_('joined at'), auto_now_add=True)

    objects = TeamManager()

    def __unicode__(self):
        return u"[%s]@%s" % (self.group, self.project)

    class Meta:
        unique_together = ('project', 'group')

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_teams_edit', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'name': self.group.name,
        })

    @models.permalink
    def get_delete_url(self):
        return ('projector_project_teams_delete', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'name': self.group.name,
        })

    @LazyProperty
    def perms(self):
        return get_perms(self.group, self.project)

class Milestone(models.Model):
    project = models.ForeignKey(Project, verbose_name=_('project'))
    name = models.CharField(max_length=64)
    slug = AutoSlugField(max_length=64, populate_from='name',
        always_update=True, unique_with='project')
    description = models.TextField()
    author = models.ForeignKey(User, verbose_name=_('author'))
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)
    deadline = models.DateField(_('deadline'), default=datetime.date.today() +
        datetime.timedelta(days=10))
            #projector_settings.get_config_value('MILESTONE_DEADLINE_DELTA')))
    date_completed = models.DateField(_('date completed'), null=True,
        blank=True)

    class Meta:
        ordering = ('created_at',)
        verbose_name = _('milestone')
        verbose_name_plural = _('milestones')
        unique_together = ('project', 'name')

    def __unicode__(self):
        return self.name

    @models.permalink
    def get_absolute_url(self):
        return ('projector_project_milestone_detail', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'milestone_slug': self.slug,
        })

    @models.permalink
    def get_edit_url(self):
        return ('projector_project_milestone_edit', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'milestone_slug': self.slug,
        })

    def get_tasks_url(self):
        _url = self.project.get_task_list_url()
        url = _url + '?milestone=%d' % self.pk
        return url

    def get_tasks_count(self):
        return self.task_set.count()

    def get_finished_tasks_count(self):
        finished_tasks = self.task_set\
            .select_related('status')\
            .filter(status__is_resolved=True)
        return finished_tasks.count()

    def get_finished_tasks_count_as_percentage(self):
        finished = self.get_finished_tasks_count()
        all = self.task_set.count()
        if all == 0:
            return 100
        return Decimal(finished) / Decimal(all) * Decimal(100)

    def get_unfinished_tasks_count(self):
        return self.task_set\
            .select_related('status')\
            .filter(status__is_resolved=False)\
            .count()

    def get_unfinished_tasks_count_as_percentage(self):
        return Decimal(100) - self.get_finished_tasks_count_as_percentage()

class TimelineEntry(models.Model):
    project = models.ForeignKey(Project, verbose_name=_('project'),
        editable=False)
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)
    user = models.ForeignKey(User, verbose_name=_('user'), null=True,
        blank=True, editable=False)
    action = models.CharField(_('action'), max_length=256,
        help_text=_('Short note on activity, i.e.: commit message'))

    class Meta:
        ordering = ('-created_at',)
        get_latest_by = 'created_at'
        verbose_name = _('timeline entry')
        verbose_name_plural = _('timeline entries')

    def __unicode__(self):
        return self.action

    def get_description(self):
        description_date_format = "%Y-%m-%d %H:%m:%s"
        formatted_date = datetime.datetime.strftime(description_date_format,
                self.created_at)
        return "%s by %s at %s" % (self.action, self.author, formatted_date)

class Priority(OrderedDictModel):
    project = models.ForeignKey(Project)
    slug = AutoSlugField(max_length=64, populate_from='name',
        always_update=True, unique_with='project')

    class Meta:
        verbose_name = _('priority level')
        verbose_name_plural = _('priority levels')
        unique_together = ('project', 'name')

class Status(OrderedDictModel):
    project = models.ForeignKey(Project)
    is_resolved = models.BooleanField(verbose_name=_('is resolved'),
        default=False)
    is_initial = models.BooleanField(verbose_name=_('is initial'),
        default=False)
    destinations = models.ManyToManyField('self', verbose_name=_('destinations'),
        through='Transition', symmetrical=False, null=True, blank=True)
    slug = AutoSlugField(max_length=64, populate_from='name',
        always_update=True, unique_with='project')

    def can_change_to(self, new_status):
        """
        Checks if ``Transition`` object with ``source`` set to ``self`` and
        ``destination`` to given ``new_status`` exists.
        """
        try:
            Transition.objects.only('id')\
                .get(source__id=self.id, destination__id=new_status.id)
            return True
        except Transition.DoesNotExist:
            return False

    class Meta:
        verbose_name = _('status')
        verbose_name_plural = _('statuses')
        unique_together = ('project', 'name')
        ordering = ['order']

    def __unicode__(self):
        return self.name

class Transition(models.Model):
    """
    Instances allow to change source Status to destination Status.
    Needed for custom workflows.
    """
    source = models.ForeignKey(Status,
        verbose_name=_('source status'),
        related_name='sources')
    destination = models.ForeignKey(Status,
        verbose_name=_('destination status'))

    class Meta:
        verbose_name = _('transition')
        verbose_name_plural = _('transitions')
        unique_together = ('source', 'destination')

    def __unicode__(self):
        return u'%s->%s' % (self.source, self.destination)

class TaskType(OrderedDictModel):
    project = models.ForeignKey(Project)

    class Meta:
        verbose_name = _('task type')
        verbose_name_plural = _('task types')
        unique_together = ('project', 'name')

class AbstractTask(models.Model):
    """
    This is base abstract class for Task and TaskRevision.
    ``created_at``, ``author`` and ``author_ip`` fields are
    used in a different way - for TaskRevision they work as
    ``edited_at``, ``editor`` and ``editor_ip`` of Task model.
    """
    created_at = models.DateTimeField(_('created at'),
        default=datetime.datetime.now, editable=False)
    author = models.ForeignKey(User, verbose_name=_('author'),
        related_name='created_%(class)s', blank=True)
    author_ip = models.IPAddressField(blank=True)
    revision = models.IntegerField(editable=False, default=0)
    summary = models.CharField(_('summary'), max_length=64)
    description = models.TextField(_('description'))
    status = models.ForeignKey(Status, verbose_name=_('status'))
    priority = models.ForeignKey(Priority, verbose_name=_('priority'))
    type = models.ForeignKey(TaskType, verbose_name=_('task type'))
    owner = models.ForeignKey(User, verbose_name=_('owner'),
        related_name='owned_%(class)s', null=True, blank=True)
    deadline = models.DateField(_('deadline'), null=True, blank=True,
        help_text='YYYY-MM-DD')
    milestone = models.ForeignKey(Milestone, verbose_name=_('milestone'),
        null=True, blank=True)
    component = models.ForeignKey(Component, verbose_name=_('component'))

    def get_status(self):
        return self.status

    def set_status(self, new_status):
        if self.status != new_status:
            self.status = new_status

    class Meta:
        abstract = True

class Task(AbstractTask, Watchable):
    id_pk = models.AutoField(primary_key=True)
    id = models.IntegerField(editable=False)
    project = models.ForeignKey(Project, verbose_name=_('project'))
    edited_at = models.DateTimeField(_('edited at'), auto_now=True)
    editor = models.ForeignKey(User, verbose_name=_('editor'), blank=True,
        null=True)
    editor_ip = models.IPAddressField(blank=True)

    objects = TaskManager()

    class Meta:
        ordering = ('-id',)
        verbose_name = _('task')
        verbose_name_plural = _('tasks')
        unique_together = ('id', 'project')

    def __unicode__(self):
        return u'#%s %s' % (self.id, self.summary)

    class CannotCalucalteIdError(Exception):
        pass

    def _calculate_id(self):
        """
        Sets and returns new id for taks within it's project.
        """
        try:
            self.project # Just check
        except Project.DoesNotExist:
            raise Task.CannotCalucalteIdError("Project is not set for this task")
        if not self.id:
            try:
                latest_task = Task.objects\
                    .filter(project=self.project)\
                    .order_by('-id')[0]
                self.id = latest_task.id + 1
            except IndexError:
                self.id = 1
        logging.debug("Task calulated id is %s" % self.id)
        return self.id

    @models.permalink
    def get_absolute_url(self):
        return ('projector_task_detail', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'task_id': self.id,
        })

    @models.permalink
    def get_edit_url(self):
        return ('projector_task_edit', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'task_id': self.id,
        })

    @models.permalink
    def get_watch_url(self):
        return ('projector_task_watch', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'task_id': self.id,
        })

    @models.permalink
    def get_unwatch_url(self):
        return ('projector_task_unwatch', (), {
            'username': self.project.author.username,
            'project_slug': self.project.slug,
            'task_id': self.id,
        })

    def save(self, *args, **kwargs):
        self._calculate_id()
        if self.pk:
            # Task update
            self.revision += 1
        return super(Task, self).save(*args, **kwargs)

    def _set_comment(self, comment):
        self._comment = comment

    def _get_comment(self):
        return getattr(self, '_comment', None)

    comment = property(_get_comment, _set_comment,
        doc = "If comment is set, clean method won't check if there were "
              "changes made to task itself")

    CHANGESET_FIELDS = (
        'summary',
        'description',
        'status',
        'component',
        'deadline',
        'milestone',
        'owner',
        'priority',
        'type',
    )

    def fetch_old(self):
        return Task.objects.get(pk=self.pk)

    @staticmethod
    def diff(new, old=None):
        """
        Returns differences between old and new task instances,
        as dict. Fields would become keys of returned dict.
        Tuples of old value and new value would be dict's values.
        """
        if old is None:
            # Fetch from database (before commit so it *is* old row)
            old = Task.objects.get(pk=new.pk)
        diff = SortedDict()
        for field in Task.CHANGESET_FIELDS:
            old_val, new_val = getattr(old, field), getattr(new, field)
            if old_val != new_val:
                diff[field] = (old_val, new_val)
        return diff

    def clean(self):
        logging.debug("Validating model... self.comment: %s" % self.comment)
        if self.id and not Task.diff(new=self) and not self.comment:
            raise ValidationError(_("No changes made"))
        return super(Task, self).clean()

    def get_revisions(self, force_query=False):
        """
        Returns TaskRevision objects related to this Task instance.
        Each TaskRevision would have additional attribute ``changes``
        which would containt result of ``Task.diff`` method. For
        first revision changes would be empty dict.
        """
        if force_query or not hasattr(self, '_revisions'):
            revisions = self.taskrevision_set\
                .select_related('type', 'author', 'status', 'editor', 'task',
                    'priority', 'milestone', 'component', 'owner',
                    'task__project')\
                .order_by('revision')
            for i, revision in enumerate(revisions):
                if i==0:
                    revision.changes = {}
                else:
                    revision.changes = Task.diff(revisions[i-1], revision)
            self._revisions = revisions
        return self._revisions

    revisions = property(get_revisions)

    def create_revision(self):
        """
        Creates revision (instance of ``TaskRevision``) for this task. If
        comment has been set, it would be used in revision creation
        process.
        """
        revision_info = dict(
            task = self,
            author = self.editor,
            author_ip = self.editor_ip,
            created_at = self.edited_at,
            revision = self.revision,
            summary = self.summary,
            description = self.description,
            owner = self.owner,
            type = self.type,
            priority = self.priority,
            status = self.status,
            milestone = self.milestone,
            component = self.component,
            deadline = self.deadline,
        )
        if self.comment is not None:
            revision_info['comment'] = self.comment
        revision = TaskRevision.objects.create(**revision_info)
        logging.debug("TaskRevision created: %s" % revision)
        if hasattr(self, '_revisions'):
            self._revisions.append(revision)
        return revision

    def get_long_summary(self):
        raw = self.project.config.task_email_summary_format
        tmpl = string.Template(raw)
        return tmpl.safe_substitute(project=self.project.name,
            id=self.id, summary=self.summary)

    def get_long_content(self):
        """
        Returns content of the task, suitable as email message's body.
        """
        task_url = 'http://%s%s' % (Site.objects.get_current().domain,
            self.get_absolute_url())
        result = render_to_string('projector/project/task/mail.html', {
            'task': self,
            'task_url': task_url,
        })
        return result

    def current_revision(self):
        """
        Returns last TaskRevision (not self) attached to this task. If there
        were no changes made yet, None is returned.
        """
        if self.revision == 0:
            return None
        else:
            return self.taskrevision_set.get(revision=self.revision)

    def notify(self, recipient_list=None):
        """
        Notifies about task's status. If ``recipient_list`` is None, would
        send a note to everyone who watch this task.
        """
        if recipient_list is None:
            recipient_list = self.get_watchers()
        if self.project.is_private():
            if isinstance(recipient_list, QuerySet):
                recipient_list = recipient_list.filter(
                    Q(membership__project=self.project) |
                    Q(project=self.project) |
                    #Q(is_superuser=True) |
                    Q(groups__team__project=self.project))
            else:
                raise TypeError("Currently only QuerySet instances are "
                    "allowed as recipient_list parameter for private "
                    "projects")
        if isinstance(recipient_list, QuerySet):
            recipient_list = recipient_list.values_list('email', flat=True)
        # Be sure not to double recipient
        recipient_list=set(recipient_list)
        mail_info = {
            'subject': self.get_long_summary(),
            'message': self.get_long_content(),
            'from_email': settings.DEFAULT_FROM_EMAIL,
            'recipient_list': [addr for addr in recipient_list],
        }
        send_mail(**mail_info)


class TaskRevision(AbstractTask):
    task = models.ForeignKey(Task)
    comment = models.TextField(_('comment'), max_length=3000,
        null=True, blank=True)

    class Meta:
        ordering = ('task', 'revision',)
        unique_together = ('revision', 'task')

    def __unicode__(self):
        return "#%d %s: Revision %d" % (self.task.id, self.task.summary,
            self.revision)

class UserProfile(RichUserProfile):
    """
    Base user profile class for ``django-projector``.
    Would be abstract if ``AUTH_PROFILE_MODULE`` is not set or doesn't equal
    with ``projector.UserProfile``.
    """
    activation_token = models.CharField(_('activation token'), max_length=32,
        editable=False)
    group = models.OneToOneField(Group, verbose_name=_('group'), null=True,
        blank=True)
    is_team = models.BooleanField(default=False)

    class Meta:
        app_label = 'projector'
        verbose_name = _('user profile')
        verbose_name_plural = _('user profiles')
        abstract = not using_projector_profile()

    def __unicode__(self):
        return u"<%s's profile>" % self.user


class UserId(models.Model):
    """
    This class would allow to identify users by some kind of id string. Created
    for repository browser - many scm's allow to use username/email combination
    and they may vary. User would be allowed to add such combinations in order
    to be identified by information given from changesets.
    If email is found within given ``raw_id``, user would need to activate this
    UserId - message would be send to the email found.
    """
    user = models.ForeignKey(User, verbose_name = _('user'))
    raw_id = models.CharField(_('Raw ID'), max_length=128, unique=True)
    is_active = models.BooleanField(_('is active'), default=False)
    activation_token = models.CharField(_('activation token'), max_length=32,
        editable=False)

    def __unicode__(self):
        return self.raw_id

    def save(self, *args, **kwargs):
        if self.raw_id:
            user = get_user_from_string(self.raw_id)
            if user:
                raise ValidationError(_("User %s has been identified by this "
                    "value" % user))
        return super(UserId, self).save(*args, **kwargs)

def get_user_from_string(value):
    """
    Returns User instance if given text can be identified or None otherwise.
    First it tries to match with User.username or User.email, next would try
    to match with UserId.raw_id.
    """
    if not value:
        return None
    try:
        return User.objects.get(username=value)
    except User.DoesNotExist:
        pass
    try:
        return User.objects.get(email=value)
    except User.DoesNotExist:
        pass
    try:
        return UserId.objects.get(raw_id=value).user
    except UserId.DoesNotExist:
        pass
    return None

# ================ #
# Signals handlers #
# ================ #

from projector.listeners import start_listening
start_listening()




from django.db.models import signals

from projector import models as projector_app
from projector.management.listeners import put_missing_project_configs,\
    update_project_permissions, create_missing_repositories

signals.post_syncdb.connect(put_missing_project_configs, sender=projector_app,
    dispatch_uid='projector.management.put_missing_project_configs')

signals.post_syncdb.connect(update_project_permissions, sender=projector_app,
    dispatch_uid='projector.management.update_project_permissions')

signals.post_syncdb.connect(create_missing_repositories, sender=projector_app,
    dispatch_uid='projector.management.create_missing_repositories')





class DefaultWorkflow(object):

    components = (
        {
            'name': u'global',
        },
    )

    task_types = (
        {
            'name': u'task',
            'order': 1,
        },
        {
            'name': u'bug',
            'order': 2,
        },
        {
            'name': u'new feature',
            'order': 3,
        },
        {
            'name': u'proposal',
            'order': 4,
        },
    )

    priorities = (
        {
            'name': u'minor',
            'order': 1,
        },
        {
            'name': u'major',
            'order': 2,
        },
        {
            'name': u'critical',
            'order': 3,
        },
        {
            'name': u'blocker',
            'order': 4,
        },
    )

    statuses = (
        {
            'name': u'new',
            'order': 1,
            'is_resolved': False,
            'is_initial': True,
        },
        {
            'name': u'assigned',
            'order': 2,
            'is_resolved': False,
            'is_initial': False,
        },
        {
            'name': u'need fix',
            'order': 3,
            'is_resolved': False,
            'is_initial': False,
        },
        {
            'name': u'need feedback',
            'order': 4,
            'is_resolved': False,
            'is_initial': False,
        },
        {
            'name': u'need unittest',
            'order': 5,
            'is_resolved': False,
            'is_initial': False,
        },
        {
            'name': u'fixed',
            'order': 6,
            'is_resolved': True,
            'is_initial': False,
        },
        {
            'name': u'closed',
            'order': 7,
            'is_resolved': True,
            'is_initial': False,
        },
        {
            'name': u'wrong',
            'order': 8,
            'is_resolved': True,
            'is_initial': False,
        },
    )




import sys

from django.conf import settings
from django.test import TestCase

class ProjectorTestCase(TestCase):

    def _get_response(self, url, data={}, method='GET', code=200, follow=False):
        """
        Test if response to given url/data/method returns with proper code.
        Returns response.
        """
        method = method.upper()
        if method == 'GET':
            opener = self.client.get
        elif method == 'POST':
            opener = self.client.post
        else:
            self.fail("Unsupported method %s" % method)

        response = opener(url, data, follow=follow)
        try:
            self.assertEqual(response.status_code, code)
        except AssertionError, err:
            answer = raw_input('\n'.join((
                "%s %s (data: %s) returned code %s but %s was expected"
                % (method, url, data, response.status_code, code),
                "AssertionError caught!",
                "USE: d (run debugger) | c (show content)",
                "[Enter to continue]",
            ))).lower()
            if answer == 'd':
                import pdb; pdb.set_trace()
            elif answer == 'c':
                sys.stdout.err.write(response.content)
            raise err

        return response

def get_homedir(project):
    """
    Returns homedir for single project.
    """
    import hashlib
    date = project.created_at.strftime('%Y%m%d%H%M%s')
    hex = hashlib.sha224(date).hexdigest()
    homedir = '-'.join(('test', date, str(project.id), hex))
    return homedir

def test_concurrently(times):
    """
    Add this decorator to small pieces of code that you want to test
    concurrently to make sure they don't raise exceptions when run at the
    same time.  E.g., some Django views that do a SELECT and then a subsequent
    INSERT might fail when the INSERT assumes that the data has not changed
    since the SELECT.

    Taken from http://www.caktusgroup.com/blog/2009/05/26/testing-django-views-for-concurrency-issues/
    """
    def test_concurrently_decorator(test_func):
        def wrapper(*args, **kwargs):
            if settings.DATABASE_ENGINE == 'sqlite3' or \
                ('default' in settings.DATABASES and
                 settings.DATABASES['default']['ENGINE'].split('.')[-1] ==
                 'sqlite3'):
                print "[WARNING] Concurrency tests are not applicable if "\
                      "sqlite3 is used as database engine"
                return
            exceptions = []
            import threading
            def call_test_func():
                try:
                    test_func(*args, **kwargs)
                except Exception, e:
                    exceptions.append(e)
                    raise
            threads = []
            for i in range(times):
                threads.append(threading.Thread(target=call_test_func))
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            if exceptions:
                raise Exception('test_concurrently intercepted %s exceptions: '
                '%s' % (len(exceptions), exceptions))
        return wrapper
    return test_concurrently_decorator




from django.db import models
from django.db import IntegrityError
from django.db.models import Q
from django.core.exceptions import ValidationError
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType

from projector.signals import setup_project

from richtemplates.shortcuts import get_first_or_None

class ProjectManager(models.Manager):

    def for_user(self, user=None):
        """
        Returns queryset of :model:`Project` instances available for given
        user. If no user is given or user is inactive/anonymous, only public
        projects are returned.
        """
        if not user:
            user = AnonymousUser()
        qs = self.get_query_set()
        qset = Q()
        if user.is_active:
            qset = qset & \
                Q(public=True) | \
                Q(public=False, membership__member=user) | \
                Q(public=False, team__group__user=user)
        else:
            qset = qset & Q(public=True)
        qs = qs.filter(qset)\
            .select_related('membership__member')\
            .order_by('name')\
            .distinct()

        return qs

    def for_member(self, user, requested_by):
        """
        Returns queryset of :model:`Project` instances which given ``user`` is
        member of with exclusion of those projects which ``requested_by`` user
        cannot see (i.e. are private and ``requested_by`` is not member of).
        """
        qs = self.for_user(requested_by)
        qset = Q()

        if user.is_anonymous():
            raise ValueError("Only requeted_by parameter may be anonymous")

        qset = qset & \
            Q(membership__member=user) | \
            Q(team__group__user=user)

        qs = qs.filter(qset)\
            .select_related('membership__member')\
            .order_by('name')\
            .distinct()

        return qs

    def create_project(self, vcs_alias=None, workflow=None, *args, **kwargs):
        """
        Creates new project and call it's setup function by sending
        :signal:`setup_project` signal.
        """
        instance = self.create(*args, **kwargs)
        setup_project.send(sender=self.model, instance=instance,
            vcs_alias=vcs_alias, workflow=workflow)
        return instance


class TaskManager(models.Manager):

    def get_for_project(self, project):
        """
        Returns Task instance with default values set related to given
        project.
        """
        from projector.models import Task
        status = get_first_or_None(
            project.status_set.filter(is_initial=True))
        type = get_first_or_None(project.tasktype_set)
        priority = get_first_or_None(project.priority_set)
        component = get_first_or_None(project.component_set)

        task = Task(
            project = project,
            status = status,
            type = type,
            priority = priority,
            component = component,
        )
        return task


class TeamManager(models.Manager):

    def for_user(self, user=None):
        """
        Returns queryset of Team instances of groups for given user.
        """
        if not user:
            user = AnonymousUser()
        queryset = self.get_query_set()
        queryset = queryset.filter(group__user=user)
        return queryset

    def convert_from_user(self, user):
        """
        Converts ``User`` instance into ``Team`` instance. It won't delete user,
        his or her profile would simply get ``team`` attribute set to newly
        created ``Team`` and ``is_team`` attribute would be set to ``True``.
        """
        from projector.models import Project
        if not user.is_active or user.is_anonymous():
            raise ValidationError("Cannot conver anonymous or inactive user")
        elif user.is_superuser or user.is_staff:
            raise ValidationError("Cannot convert staff member or superuser")
        try:
            group = Group.objects.create(name=user.username)
            profile = user.get_profile()
            profile.is_team = True
            profile.group = group
            profile.save()
            user.groups.add(group)
            for project in Project.objects.filter(author=user):
                self.create(project = project, group = group)
            return group
        except IntegrityError:
            raise ValidationError("Cannot convert user if a group with same "
                "name already exist")


class WatchedItemManager(models.Manager):

    def projects_for_user(self, user):
        """
        Returns Project instances watched by user.
        """
        from projector.models import Project
        qs = self.get_query_set().filter(
            content_type = ContentType.objects.get_for_model(Project),
            user = user,
        )
        return qs

    def task_for_user(self, user):
        """
        Returns Task instances watched by user.
        """
        from projector.models import Task
        qs = self.get_query_set().filter(
            content_type = ContentType.objects.get_for_model(Task),
            user = user,
        )
        return qs




from django.core import mail
from django.contrib.auth.models import User, Group
from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse

from projector.models import Task, Team, Membership

class EmailTest(TestCase):

    fixtures = ['test_data.json']

    def setUp(self):
        self.client = Client()

    def _create_task(self, user, project, summary='Task summary',
        description='Example description', watch_changes=True):
        """
        Creates task for given parameters. Assumes given user is already
        logged in by ``self.client``.

        Returns created task.
        """
        task = Task.objects.get_for_project(project)
        data = {
            'summary': summary,
            'description': description,
            'status': task.status_id,
            'priority': task.priority_id,
            'type': task.type_id,
            'owner': user.username,
            'component': task.component_id,
        }
        if watch_changes:
            data['watch_changes'] = u'on'
        response = self.client.post(reverse('projector_task_create', kwargs={
            'username': user.username,
            'project_slug': project.slug}),
            data=data,
            follow=True)
        self.assertEqual(response.status_code, 200)
        return user.task_set.filter(project=project).order_by('-created_at')[0]

    def test_tasks_notification(self):
        # needs project to be created first
        jack = User.objects.get(username='jack')

        self.client.login(username='jack', password='jack')
        project = jack.project_set.all()[0]

        task = self._create_task(user=jack, project=project)

        # Author of the project should be notified after task is created
        self.assertEquals(len(mail.outbox), 1)

        joe = User.objects.get(username='joe')
        self.client.login(username='joe', password='joe')
        task.watch(joe)

        # After joe started to watch task, calling ``notify`` should send one
        # mail with both jack and joe in recipient list
        task.notify()
        self.assertEquals(len(mail.outbox), 2)
        self.assertEquals(
            set([jack.email, joe.email]),
            set(mail.outbox[1].recipients())
        )

        # If we set recipient list arbitrary we expect only those recipients
        # would receive the message
        recipient_list = [joe.email]
        task.notify(recipient_list=recipient_list)
        self.assertEquals(len(mail.outbox), 3)
        self.assertEquals(recipient_list, mail.outbox[2].recipients())

        # If project is private we need to ensure only members and author/owner
        # would receive the message
        task.project.public = False
        task.project.save()
        task.notify()
        self.assertEquals(len(mail.outbox), 4)
        self.assertEquals([jack.email], mail.outbox[3].recipients())

        # But if joe would join Team associated with task's project then he
        # could get the message
        group = Group.objects.create(name='jack-n-joe')
        joe.groups.add(group)
        Team.objects.create(group=group, project=task.project)
        task.notify()
        self.assertEquals(len(mail.outbox), 5)
        self.assertEquals(
            set([jack.email, joe.email]),
            set(mail.outbox[4].recipients()),
        )

        # Membership would be enough, too
        joe.groups.remove(group)
        Membership.objects.create(project=task.project, member=joe)
        task.notify()
        self.assertEquals(len(mail.outbox), 6)
        self.assertEquals(
            set([jack.email, joe.email]),
            set(mail.outbox[5].recipients()),
        )




"""
Tests here shouldn't be run with sqlite backend.
"""
from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse

from projector.tests.base import test_concurrently

class ThreadedProjectTest(TestCase):

    def test_list(self):

        url = reverse('projector_project_list')

        @test_concurrently(10)
        def toggle_test():
            client = Client()
            client.get(url)

        toggle_test()




from django.test import TestCase
from django.test.client import Client
from django.contrib.auth.models import User, AnonymousUser
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest

from projector.core.exceptions import ProjectorError, ForkError
from projector.models import Project
from projector.tests.base import ProjectorTestCase
from projector.forks.base import BaseExternalForkForm
from projector.forks.bitbucket import BitbucketForkForm

class ForkTest(TestCase):

    def setUp(self):
        self.anon = AnonymousUser()
        self.joe = User.objects.create(username='joe')
        self.project = Project.objects.create_project(name='project',
            author=self.joe)


    def test_anonymous(self):
        """
        AnonymousUser cannot fork projects.
        """

        self.assertRaises(PermissionDenied, self.project.fork, user=self.anon)

    def test_inactive(self):
        """
        Inactive users cannot fork projects.
        """
        jack = User.objects.create(username='jack', is_active=False)
        self.assertRaises(PermissionDenied, self.project.fork, user=jack)

    def test_not_viewable(self):
        """
        If user cannot view project he/she cannot fork it either.
        """
        Project.objects.filter(id=self.project.id).update(public=False)
        jack = User.objects.create(username='jack')
        self.assertRaises(PermissionDenied, self.project.fork, user=jack)

    def test_same_author(self):
        """
        User cannot fork his/her own project.
        """
        self.assertRaises(ForkError, self.project.fork, user=self.project.author)

    def test_already_forked(self):
        """
        User cannot fork same project twice. Tests ``get_fork_for_user`` method
        too.
        """
        jack = User.objects.create(username='jack')
        jack_fork = self.project.fork(user=jack)
        f = self.project.get_fork_for_user(jack)
        self.assertTrue(jack_fork.id == f.id,
            "fork method returned %s (id=%s) and get_fork_for_user returned "
            "%s (id=%s)" % (jack_fork, jack_fork.id, f, f.id))

        jacky = User.objects.create(username='jacky')
        jacky_fork = self.project.fork(user=jacky)
        f = self.project.get_fork_for_user(jacky)
        self.assertTrue(jacky_fork.id == f.id,
            "fork method returned %s (id=%s) and get_fork_for_user returned "
            "%s (id=%s)" % (jacky_fork, jacky_fork.id, f, f.id))

        jade = User.objects.create(username='jade')
        jade_fork = jack_fork.fork(user=jade)
        f = self.project.get_fork_for_user(jade)
        self.assertTrue(jade_fork.id == f.id,
            "fork method returned %s (id=%s) and get_fork_for_user returned "
            "%s (id=%s)" % (jack_fork, jack_fork.id, f, f.id))
        jacky_fork.name = 'changed-name'
        self.assertRaises(ForkError, jacky_fork.fork, user=jade)

    def test_force_private(self):
        jack = User.objects.create(username='jack')
        fork = self.project.fork(user=jack, force_private=True)
        self.assertFalse(fork.public)


class ForkViewTest(ProjectorTestCase):

    def setUp(self):
        self.joe = User.objects.create_user(username='joe',
            email='joe@example.com', password='joe')
        self.project = Project.objects.create_project(name='project',
            author=self.joe)
        self.jack = User.objects.create_user(username='jack',
            email='jack@example.com', password='jack')
        self.client = Client()

    def test_fork(self):
        self.assertFalse(self.project.is_fork())
        self.client.login(username='jack', password='jack')

        # Jack goes to joe's project page
        url = self.project.get_absolute_url()
        response = self._get_response(url)
        self.assertEqual(response.context['user_fork'], None)
        self.assertEqual(response.context['forks'],
            [response.context['project']])

        # Jack forks joe's project
        url = self.project.get_fork_url()
        response = self._get_response(url, {
            'submit': u'Fork project'},
            method='POST', follow=True)
        fork = response.context['project']
        self.assertTrue(fork.is_fork())
        self.assertEqual(fork.get_root().id, self.project.id)

        if self.project.repository is not None:
            self.assertEqual(fork.repository.alias,
                self.project.repository.alias,
                "Alias of fork (%s) should be same as alias of original "
                "project (%s)" % (fork, self.project))
            self.assertEqual(fork.repository.revisions,
                self.project.repository.revisions)


class BaseExternalForkFormTest(TestCase):

    def setUp(self):
        self.request = HttpRequest()

    def test_not_cleaned_no_data(self):
        form = BaseExternalForkForm()
        form.request = self.request
        self.assertRaises(ProjectorError, form.is_public)
        self.assertRaises(ProjectorError, form.is_private)

    def test_not_cleaned_with_data(self):
        form = BaseExternalForkForm({'as_private': u'checked'})
        form.request = self.request
        self.assertRaises(ProjectorError, form.is_public)
        self.assertRaises(ProjectorError, form.is_private)

    def test_valid_public(self):
        form = BaseExternalForkForm({})
        form.request = self.request
        self.assertTrue(form.is_valid())
        # as_private not checked
        self.assertTrue(form.is_public())
        self.assertFalse(form.is_private())

    def test_valid_private(self):
        form = BaseExternalForkForm({'as_private': u'checked'})
        form.request = self.request
        self.assertTrue(form.is_valid())
        # as_public checked
        self.assertFalse(form.is_public())
        self.assertTrue(form.is_private())


class BitbucketForkTest(TestCase):

    def setUp(self):
        self.request = HttpRequest()
        self.request.user = User.objects.create(username='joe')

    def test_fork(self):
        data = {
            'username': u'marcinkuzminski',
            'projectname': u'vcs',
            'as_private': u'checked',
        }
        form = BitbucketForkForm(data)
        form.request = self.request
        self.assertTrue(form.is_valid())
        fork = form.fork()
        fork = Project.objects.get(pk=fork.pk)
        self.assertTrue(len(fork.repository.revisions) > 100)
        self.assertTrue(fork.is_private())

    def test_fork_with_same_name(self):
        Project.objects.create(name=u'joe-project', author=self.request.user)
        data = {
            'username': u'whatever',
            'projectname': u'joe-project',
            'as_private': u'checked',
        }
        form = BitbucketForkForm(data)
        form.request = self.request
        self.assertFalse(form.is_valid())

    # Deprecated: errors are handled during project setup as projects are now
    # created asynchronously

    #def test_wrong_values(self):
        #joe = User.objects.create(username='joe')
        #data = {
            #'username': u'x',
            #'projectname': u'x',
        #}
        #form = BitbucketForkForm(data)
        #self.assertTrue(form.is_valid())
        #request = HttpRequest()
        #request.user = joe
        #try:
            #form.fork(request)
        #except ProjectorError:
            #pass
        #else:
            #self.fail("Form should raise subclass of ProjectorError")

    def test_valid(self):
        """
        This test checks if sane values have been passed to the form.
        Allowing external forking may be very dangerous as we may expose
        own project to be used as *proxy* for attacks on external locations.
        """
        data_list = [dict((key, val) for key, val in (
            ('foobar', '<script...'),
            ('<script', 'foobar'),
            ('foobar', '../../'),
            ('../../', 'foobar'),
        ))]
        for data in data_list:
            form = BitbucketForkForm(data)
            form.request = self.request
            self.assertFalse(form.is_valid())




from django.test import TestCase
from django.http import HttpRequest, HttpResponse
from django.contrib.auth.models import AnonymousUser

from projector.core.controllers import View

class ViewTest(TestCase):

    def setUp(self):
        self.request = HttpRequest()
        self.request.user = AnonymousUser()

    def test_response(self):

        class AnyView(View):
            template_name = 'tests/dummy.html'

        response = AnyView(self.request)
        self.assertTrue(isinstance(response, HttpResponse))

    def test_dict_response(self):

        class DictView(View):
            template_name = 'tests/dummy.html'
            def response(self, request):
                return self.context

        response = DictView(self.request)
        self.assertTrue(isinstance(response, HttpResponse))

    def test_login_required(self):

        class LoginRequiredView(View):
            template_name = 'tests/dummy.html'
            login_required = True
            def response(self, request):
                return self.context

        self.request.user = AnonymousUser()
        response = LoginRequiredView(self.request)
        self.assertEqual(response.status_code, 302)

    def test_before(self):

        test = self

        class BeforeView(View):
            template_name = 'tests/dummy.html'
            def __before__(self):
                self.context['before'] = 'before'
            def response(self, request):
                test.assertEqual(self.context['before'], 'before')

        self = test
        # Assertion is called inside response method at the BeforeView
        BeforeView(self.request)

    def test_before_response(self):

        class BeforeResponseView(View):
            template_name = 'tests/dummy.html'
            def __before__(self):
                return HttpResponse(content='foobar', status=500)
            def response(self, request):
                return HttpResponse(status=200)

        response = BeforeResponseView(self.request)
        self.assertEqual(response.status_code, 500)

    def test_after(self):

        test = self

        class AfterView(View):
            template_name = 'tests/dummy.html'
            def __after__(self):
                self.context['after'] = 'after'
                if self._response.status_code == 200:
                    self._response.status_code = 501
            def response(self, request):
                test.assertFalse('after' in self.context)
                return self.context

        self = test
        # Assertion is called inside response method at AfterView
        AfterView(self.request)

    def test_after_response(self):

        class AfterResponseView(View):
            template_name = 'tests/dummy.html'
            def __after__(self):
                return HttpResponse(content='foobar', status=500)
            def response(self, request):
                return HttpResponse(status=200)

        response = AfterResponseView(self.request)
        self.assertEqual(response.status_code, 500)

    def test_before_after_mix(self):

        test = self

        class BeforeAndAfterView(View):
            template_name = 'tests/dummy.html'
            def __before__(self):
                self.context['before'] = 'before'
                self.context['mix'] = 'before'
            def __after__(self):
                self.context['after'] = 'after'
                self.context['mix'] = 'after'
            def response(self, request):
                test.assertEqual(self.context['before'], 'before')
                test.assertEqual(self.context['mix'], 'before')
                test.assertFalse('after' in self.context)
                return self.context

        self = test
        # Assertions are called inside response method at BeforeAndAfterView
        BeforeAndAfterView(self.request)

